from .getcapabilities import WCS20GetCapabilitiesHandler
from .describecoverage import WCS20DescribeCoverageHandler
from .getcoverage import WCS20GetCoverageHandler
from .describeeocoverageset import WCS20DescribeEOCoverageSetHandler
from .geteocoverageset import WCS20GetEOCoverageSetHandler


GetCapabilitiesHandler = WCS20GetCapabilitiesHandler
DescribeCoverageHandler = WCS20DescribeCoverageHandler
DescribeEOCoverageSetHandler = WCS20DescribeEOCoverageSetHandler
GetEOCoverageSetHandler = WCS20GetEOCoverageSetHandler
GetCoverageHandler = WCS20GetCoverageHandler



#-------------------------------------------------------------------------------
#
#  Response form objects.
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

try:
    # available in Python 2.7+
    from collections import OrderedDict
except ImportError:
    from django.utils.datastructures import SortedDict as OrderedDict
from django.utils.six import iteritems

from .base import ParamMetadata


class Output(ParamMetadata):
    # pylint: disable=too-few-public-methods,too-many-arguments
    """ Requested output definition.

    Constructor parameters:
        identifier   output identifier
        title        output title (human-readable name)
        abstract     output abstract (human-readable description)
        uom          output LiteralData UOM
        crs          output BoundingBox CRS
        mime_type    output ComplexData mime-type
        encoding     output ComplexData encoding
        schema       output ComplexData schema
        as_reference  boolean flag indicating whether the output should
                     passed as a reference op directly in the response.
    """
    def __init__(self, identifier, title=None, abstract=None, uom=None,
                 crs=None, mime_type=None, encoding=None, schema=None,
                 as_reference=False):
        ParamMetadata.__init__(
            self, identifier, title, abstract, uom, crs,
            mime_type, encoding, schema
        )
        self.as_reference = as_reference


class ResponseForm(OrderedDict):
    """ Response form defined as an ordered dictionary of the output
    definitions.
    """

    def __init__(self):
        super(ResponseForm, self).__init__()

    def set_output(self, output):
        """ Set (insert) a new definition output."""
        self[output.identifier] = output

    def get_output(self, identifier):
        """ Get an output for the given output identifier.
            An instance of the Output object is always returned.
        """
        output = self.get(identifier, None)
        if not output:
            output = Output(identifier)
        return output


class ResponseDocument(ResponseForm):
    """ Object representation of the (WPS Execute) response document.

    Constructor parameters (meaning described in  OGC 05-007r7, Table 50):
        lineage         boolean flag, set to True to print the lineage
        status          boolean flag, set to True to update status
        store_response  boolean flag, set to True to store execute response
    """
    raw = False

    def __init__(self, lineage=False, status=False, store_response=False):
        super(ResponseDocument, self).__init__()
        self.lineage = lineage
        self.status = status
        self.store_response = store_response

    def __reduce__(self): # NOTE: needed for correct async-WPS request pickling
        return (
            self.__class__, (self.lineage, self.status, self.store_response),
            None, None, iteritems(self)
        )

    def __str__(self):
        return (
            "ResponseDocument(lineage=%s, status=%s, store_response=%s)[%s]" %
            (self.lineage, self.status, self.store_response, " ,".join(
                repr(k) for k in self.keys()
            ))
        )


class RawDataOutput(ResponseForm):
    """ Object representation of the raw output response.

    Constructor parameters:
        output      name of the requested output parameter
    """
    raw = True
    lineage = False
    status = False
    store_response = False

    def __init__(self, output):
        super(RawDataOutput, self).__init__()
        self.set_output(output)

    def __str__(self):
        return (
            "RawDataOutput()[%s]" % " ,".join(repr(k) for k in self.keys())
        )



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2019 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from eoxserver.core.util.xmltools import parse
from eoxserver.core.decoders import xml, to_dict, InvalidParameterException

from eoxserver.resources.coverages.metadata.utils.gsc import (
    NS_GSC, nsmap, GSCFormatDecoder
)


class GSCFormatExtendedDecoder(xml.Decoder):
    namespaces = nsmap

    identifier = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:metaDataProperty/gsc:EarthObservationMetaData/eop:identifier/text()", type=str, num="?")
    parent_identifier = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:metaDataProperty/gsc:EarthObservationMetaData/eop:parentIdentifier/text()", type=str, num="?")
    production_status = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:metaDataProperty/gsc:EarthObservationMetaData/eop:status/text()", type=str, num="?")
    acquisition_type = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:metaDataProperty/gsc:EarthObservationMetaData/eop:acquisitionType/text()", type=str, num="?") 
    orbit_number = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:using/eop:EarthObservationEquipment/eop:acquisitionParameters/opt:Acquisition/eop:orbitNumber/text()", type=int, num="?")
    orbit_direction = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:using/eop:EarthObservationEquipment/eop:acquisitionParameters/opt:Acquisition/eop:orbitDirection/text()", type=str, num="?")
    
    swath_identifier = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:using/eop:EarthObservationEquipment/eop:sensor/eop:Sensor/eop:swathIdentifier/text()", type=str, num="?")
   

    product_version = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:resultOf/eop:EarthObservationResult/eop:product/eop:ProductInformation/eop:version/text()", type=float, num="?")
    product_quality_status = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:resultOf/eop:EarthObservationMetaData/eop:productQualityStatus/text()", type=str, num="?")
    product_quality_degradation_tag = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:metaDataProperty/gsc:EarthObservationMetaData/eop:productQualityDegradationTag/text()", type=str, num="?")
    processing_center = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:metaDataProperty/gsc:EarthObservationMetaData/eop:processing/eop:ProcessingInformation/eop:processingCenter/text()", type=str, num="?")
    creation_date = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:metaDataProperty/gsc:EarthObservationMetaData/eop:creationDate/text()", type=str, num="?") # insertion into catalog
    modification_date = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:metaDataProperty/gsc:EarthObservationMetaData/eop:modificationDate/text()", type=str, num="?") # last modification in catalog
    processing_date = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:metaDataProperty/gsc:EarthObservationMetaData/eop:processing/eop:ProcessingInformation/eop:processingDate/text()", type=str, num="?")
    sensor_mode = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:using/eop:EarthObservationEquipment/eop:sensor/eop:Sensor/eop:operationalMode/text()", type=str, num="?")
    archiving_center = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:metaDataProperty/gsc:EarthObservationMetaData/eop:archivedIn/eop:ArchivingInformation/eop:archivingCenter/text()", type=str, num="?")
    processing_mode = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:metaDataProperty/gsc:EarthObservationMetaData/eop:processing/eop:ProcessingInformation/eop:processingMode/text()", type=str, num="?")
    # acquisition type metadata
    
    acquisition_station = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:metaDataProperty/gsc:EarthObservationMetaData/eop:downlinkedTo/eop:DownlinkInformation/eop:acquisitionStation/text()", type=str, num="?")
    acquisition_sub_type = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:metaDataProperty/gsc:EarthObservationMetaData/eop:acquisitionSubType/text()", type=str, num="?")
    start_time_from_ascending_node = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:using/eop:EarthObservationEquipment/eop:acquisitionParameters/eop:Acquisition/eop:startTimeFromAscendingNode/text()", type=str, num="?")
    completion_time_from_ascending_node = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:using/eop:EarthObservationEquipment/eop:acquisitionParameters/eop:Acquisition/eop:completionTimeFromAscendingNode/text()", type=str, num="?")
    illumination_azimuth_angle = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:using/eop:EarthObservationEquipment/eop:acquisitionParameters/opt:Acquisition/opt:illuminationAzimuthAngle/text()", type=float, num="?")
    illumination_zenith_angle = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:using/eop:EarthObservationEquipment/eop:acquisitionParameters/opt:Acquisition/opt:illuminationZenithAngle/text()", type=float, num="?")
    illumination_elevation_angle = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:using/eop:EarthObservationEquipment/eop:acquisitionParameters/opt:Acquisition/opt:illuminationElevationAngle/text()", type=float, num="?")
    across_track_incidence_angle = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:using/eop:EarthObservationEquipment/eop:acquisitionParameters/opt:Acquisition/eop:acrossTrackIncidenceAngle/text()", type=float, num="?")
    along_track_incidence_angle = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:using/eop:EarthObservationEquipment/eop:acquisitionParameters/opt:Acquisition/eop:alongTrackIncidenceAngle/text()", type=float, num="?")
    polarisation_mode = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:using/eop:EarthObservationEquipment/eop:acquisitionParameters/sar:Acquisition/sar:polarisationMode/text()", type=str, num="?")
    polarization_channels = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:using/eop:EarthObservationEquipment/eop:acquisitionParameters/sar:Acquisition/sar:polarisationChannels/text()", type=str, num="?")
    antenna_look_direction = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:using/eop:EarthObservationEquipment/eop:acquisitionParameters/sar:Acquisition/sar:antennaLookDirection/text()", type=str, num="?")
    minimum_incidence_angle = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:using/eop:EarthObservationEquipment/eop:acquisitionParameters/sar:Acquisition/sar:minimumIncidenceAngle/text()", type=float, num="?")
    maximum_incidence_angle = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:using/eop:EarthObservationEquipment/eop:acquisitionParameters/sar:Acquisition/sar:maximumIncidenceAngle/text()", type=float, num="?")
    # for SAR acquisitions
    doppler_frequency = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:using/eop:EarthObservationEquipment/eop:acquisitionParameters/sar:Acquisition/sar:dopplerFrequency/text()", type=float, num="?")
    incidence_angle_variation = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:using/eop:EarthObservationEquipment/eop:acquisitionParameters/sar:Acquisition/sar:incidenceAngleVariation/text()", type=float, num="?")
    # for OPT/ALT
    cloud_cover = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:resultOf/opt:EarthObservationResult/opt:cloudCoverPercentage/text()", type=float, num="?")
    snow_cover = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:resultOf/opt:EarthObservationResult/opt:snowCoverPercentage/text()", type=float, num="?")
    
    #TODO
    # add track and frame 
    # add the following values:
    # availability_time = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/om:resultTime/gml:TimeInstant/ gml:timePosition/text()", type=str, num="?")
    # lowest_location = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:using/lmb:Footprint/lmb:minimumAltitude/text()", type=float, num= "?")
    # highest_location = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:using/lmb:Footprint/lmb:maximumAltitude/text()", type=float, num= "?")

class GSCProductMetadataReader(object):
    def test(self, obj):
        tree = parse(obj)

        tag = tree.getroot().tag if tree is not None else None
        return tree is not None and tag == NS_GSC('report')

    def read(self, obj):
        tree = parse(obj)
        if tree is not None:
            decoder = GSCFormatDecoder(tree)
            values = {
                "identifier": decoder.identifier,
                "begin_time": decoder.begin_time,
                "end_time": decoder.end_time,
                "footprint": decoder.footprint,
                "format": "gsc",
            }
            values.update(to_dict(GSCFormatExtendedDecoder(tree)))
            return values



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from django.db.models import Q

from eoxserver.core import Component
from eoxserver.core.config import get_eoxserver_config
from eoxserver.core.decoders import config, typelist
from eoxserver.contrib import mapserver as ms
from eoxserver.contrib import gdal
from eoxserver.render.coverage.objects import Coverage
from eoxserver.resources.coverages import crss, models
# from eoxserver.resources.coverages.models import RectifiedStitchedMosaic
from eoxserver.resources.coverages.formats import getFormatRegistry


class WCSConfigReader(config.Reader):
    section = "services.ows.wcs"
    supported_formats = config.Option(type=typelist(str, ","), default=())
    maxsize = config.Option(type=int, default=None)

    section = "services.ows"
    update_sequence = config.Option(default="0")


class BaseRenderer(Component):
    abstract = True

    def create_map(self):
        """ Helper function to create a WCS enabled MapServer mapObj.
        """
        map_ = ms.mapObj()
        map_.setMetaData("ows_enable_request", "*")
        map_.setMetaData("ows_onlineresource", "*")
        maxsize = WCSConfigReader(get_eoxserver_config()).maxsize
        if maxsize is not None:
            map_.maxsize = maxsize
        map_.setMetaData("ows_updateSequence",
            WCSConfigReader(get_eoxserver_config()).update_sequence
        )
        return map_

    def arraydata_locations_for_coverage(self, coverage):
        """ Helper function to query all relevant data items for any raster data
            from the database.
        """
        if isinstance(coverage, Coverage):
            return coverage.arraydata_locations
        return []

    def layer_for_coverage(self, coverage, native_format, version=None):
        """ Helper method to generate a WCS enabled MapServer layer for a given
            coverage.
        """
        range_type = coverage.range_type
        bands = list(range_type)

        coverage_identifier = coverage.identifier
        if version and version.startswith('2'):
            try:
                models.identifier_validators[0](coverage_identifier)
            except Exception:
                coverage_identifier = 'not-ncname'

        # create and configure layer
        layer = ms.layerObj()
        layer.name = coverage_identifier
        layer.type = ms.MS_LAYER_RASTER

        sr = coverage.grid.spatial_reference

        layer.setProjection(sr.proj)

        extent = coverage.extent
        size = coverage.size
        resolution = ((extent[2] - extent[0]) / float(size[0]),
                      (extent[1] - extent[3]) / float(size[1]))

        layer.setExtent(*extent)

        ms.setMetaData(layer, {
            "title": coverage_identifier,
            "enable_request": "*"
        }, namespace="ows")

        data_type = bands[0].data_type

        if bands[0].allowed_values:
            interval = bands[0].allowed_values[0]
        else:
            interval = gdal.GDT_NUMERIC_LIMITS[data_type]

        if bands[0].significant_figures is not None:
            significant_figures = bands[0].significant_figures
        else:
            significant_figures = gdal.GDT_SIGNIFICANT_FIGURES[data_type]

        ms.setMetaData(layer, {
            "label": coverage_identifier,
            "extent": "%.10g %.10g %.10g %.10g" % extent,
            "resolution": "%.10g %.10g" % resolution,
            "size": "%d %d" % size,
            "bandcount": str(len(bands)),
            "interval": "%f %f" % interval,
            "significant_figures": "%d" % significant_figures,
            "rangeset_name": range_type.name,
            "rangeset_label": range_type.name,
            "imagemode": ms.gdalconst_to_imagemode_string(bands[0].data_type),
            "formats": " ".join([
                f.wcs10name if version.startswith("1.0") else f.mimeType
                for f in self.get_wcs_formats()]
            )
        }, namespace="wcs")

        if version is None or version.startswith("2.0"):
            ms.setMetaData(layer, {
                "band_names": " ".join([band.identifier for band in bands]),
            }, namespace="wcs")
        else:
            ms.setMetaData(layer, {
                "rangeset_axes": ",".join(band.identifier for band in bands),
            }, namespace="wcs")

        if native_format:
            if version.startswith("1.0"):
                native_format = next((
                    x.wcs10name for x in self.get_wcs_formats()
                    if x.mimeType == native_format), native_format
                )
            ms.setMetaData(layer, {
                "native_format": native_format,
                "nativeformat": native_format
            }, namespace="wcs")

        native_crs = "EPSG:%d" % sr.srid
        all_crss = list(crss.getSupportedCRS_WCS(format_function=crss.asShortCode))
        if native_crs in all_crss:
            all_crss.remove(native_crs)

        # setting the coverages CRS as the first one is important!
        all_crss.insert(0, native_crs)

        supported_crss = " ".join(all_crss)
        layer.setMetaData("ows_srs", supported_crss)
        layer.setMetaData("wcs_srs", supported_crss)

        for band in bands:
            ms.setMetaData(layer, {
                "band_description": band.description,
                "band_definition": band.definition,
                "band_uom": band.unit_of_measure,
            }, namespace=band.identifier)

            if band.allowed_values:
                interval = band.allowed_values[0]
            else:
                interval = gdal.GDT_NUMERIC_LIMITS[band.data_type]

            # For MS WCS 1.x interface
            ms.setMetaData(layer, {
                "label": band.identifier,
                "interval": "%d %d" % interval
            }, namespace="wcs_%s" % band.identifier)

        if bands[0].nil_values:
            nilvalues, nilvalues_reasons = zip(*[
                [nv[0], nv[1]] for nv in bands[0].nil_values]
            )
            if nilvalues:
                ms.setMetaData(layer, {
                    "nilvalues": " ".join(nilvalues),
                    "nilvalues_reasons": " ".join(nilvalues_reasons)
                }, namespace="wcs")

        return layer

    def get_native_format(self, coverage, data_locations):
        formats = set(
            data_location.format
            for data_location in data_locations
            if data_location.format
        )
        if len(formats) == 1:
            return formats.pop()

        return None

    def find_param(self, params, name, default=None):
        for key, value in params:
            if key == name:
                return value
        return default

    def get_wcs_formats(self):
        return getFormatRegistry().getSupportedFormatsWCS()

    def get_all_outputformats(self, use_mime=True):
        outputformats = []
        for frmt in self.get_wcs_formats():
            of = ms.outputFormatObj(frmt.driver, "custom")
            of.name = frmt.mimeType if use_mime else frmt.wcs10name
            of.mimetype = frmt.mimeType
            of.extension = frmt.defaultExt
            outputformats.append(of)
        return outputformats


def is_format_supported(mime_type):
    reader = WCSConfigReader(get_eoxserver_config())
    return mime_type in reader.supported_formats



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------


from lxml.builder import E, ElementMaker

from eoxserver.core.util.xmltools import XMLEncoder, NameSpace, NameSpaceMap

ns_xlink = NameSpace("http://www.w3.org/1999/xlink", "xlink")
nsmap = NameSpaceMap(ns_xlink)
E_WITH_XLINK = ElementMaker(nsmap=nsmap)


class WMS11Encoder(XMLEncoder):
    def encode_capabilities(self, config, ows_url, srss, formats, info_formats,
                            layer_descriptions):
        return E("WMT_MS_Capabilities",
            E("Service",
                E("Name", config.name),
                E("Title", config.title),
                E("Abstract", config.abstract),
                E("KeywordList", *[
                    E("Keyword", keyword)
                    for keyword in config.keywords
                ]),
                E("OnlineResource", config.onlineresource),

                E("ContactInformation",
                    E("ContactPersonPrimary",
                        E("ContactPerson", config.individual_name),
                        E("ContactOrganization",  config.provider_name),
                    ),
                    E("ContactPosition", config.position_name),
                    E("ContactAddress",
                        E("AddressType", "postal"),
                        E("Address", config.delivery_point),
                        E("City", config.city),
                        E("StateOrProvince", config.administrative_area),
                        E("PostCode", config.postal_code),
                        E("Country", config.country),
                    ),
                    E("ContactVoiceTelephone", config.phone_voice),
                    E("ContactFacsimileTelephone", config.phone_facsimile),
                    E("ContactElectronicMailAddress",
                        config.electronic_mail_address
                    ),
                ),
                E("Fees", config.fees),
                E("AccessConstraints", config.access_constraints),
            ),
            E("Capability",
                E("Request",
                    E("GetCapabilities",
                        E("Format", "application/vnd.ogc.wms_xml"),
                        self.encode_dcptype(ows_url)
                    ),
                    E("GetMap", *[
                            E("Format", frmt.mimeType)
                            for frmt in formats
                        ] + [
                            self.encode_dcptype(ows_url)
                        ]
                    ),
                    E("GetFeatureInfo",
                        E("Format",
                            # TODO
                        ),
                        self.encode_dcptype(ows_url)
                    ),
                    # TODO: describe layer?
                ),
                E("Exception",
                    E("Format", "application/vnd.ogc.se_xml"),
                    E("Format", "application/vnd.ogc.se_inimage"),
                    E("Format", "application/vnd.ogc.se_blank"),
                ),
                E("Layer",
                    E("Title", config.title),
                    E("LatLonBoundingBox",
                        minx="-180", miny="-90", maxx="180", maxy="90"
                    ), *([
                        E("SRS", srs)
                        for srs in srss
                    ] + [
                        self.encode_layer(layer_description)
                        for layer_description in layer_descriptions
                    ])
                )
            ),
            version="1.1.1", updateSequence=config.update_sequence
        )

    def encode_dcptype(self, ows_url):
        return E("DCPType",
            E("HTTP",
                E("Get",
                    E_WITH_XLINK("OnlineResource", **{
                        ns_xlink("href"): ows_url,
                        ns_xlink("type"): "simple"
                    })
                )
            )
        )

    def encode_layer(self, layer_description):
        elems = [
            E("Name", layer_description.name)
        ]

        title = getattr(layer_description, 'title')
        if title:
            elems.append(E("Title", title))

        if layer_description.bbox:
            bbox = map(str, layer_description.bbox)
            elems.append(
                E("LatLonBoundingBox",
                    minx=bbox[0], miny=bbox[1], maxx=bbox[2], maxy=bbox[3]
                )
            )

        elems.extend(
            E("Style",
                E("Name", style),
                E("Abstract", style),
            ) for style in layer_description.styles
        )

        elems.extend(
            self.encode_layer(sub_layer)
            for sub_layer in layer_description.sub_layers
        )

        dimensions = []
        extents = []

        for dimension_name, dimension in layer_description.dimensions.items():
            dimension_elem = E("Dimension", name=dimension_name)
            if "units" in dimension:
                dimension_elem.attrib["units"] = dimension["units"]
            dimensions.append(dimension_elem)

            if "min" in dimension and "max" in dimension and "step" in dimension:
                extent_text = "%s/%s/%s" % (
                    dimension["min"], dimension["max"], dimension["step"]
                )
            elif "values" in dimension:
                extent_text = ",".join(dimension["values"])

            extent_elem = E("Extent", extent_text, name=dimension_name)

            if "default" in dimension:
                extent_elem.attrib["default"] = dimension["default"]

            if dimension.get("multivalue"):
                extent_elem.attrib["multipleValues"] = "1"

            extents.append(extent_elem)

        elems.extend(dimensions)
        elems.extend(extents)

        return E("Layer",
            *elems,
            queryable="1" if layer_description.queryable else "0"
        )



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import time
import logging


_logger = logging.getLogger(__name__)


class DurationMeasurement(object):
    def __init__(self, name, logger, level):
        self.name = name
        self.logger = logger
        self.level = level
        self.start = None
        self.end = None

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, *args, **kwargs):
        self.end = time.time()
        msg = "'%s' took %f seconds." % (self.name, self.duration)
        self.logger.log(self.level, msg)

    @property
    def duration(self):
        if self.start is not None and self.end is not None:
            return self.end - self.start
        return None


def log_duration(name, logger=None, level=logging.DEBUG):
    """ Convenience function to log the duration of a specific event.

        :param name: The name of the event.
        :param logger: The logger to use.
        :param level: The log level to log the final message to.
    """

    logger = logger or _logger
    return DurationMeasurement(name, logger, level)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


class RenderParameters(object):
    """ Abstract base class for render parameters
    """

    def __iter__(self):
        """ Yields all kvps as key-value pairs (string, string).
        """


class VersionedParams(object):
    def __init__(self, version):
        self._version = version

    @property
    def version(self):
        return self._version

    def __iter__(self):
        yield ("version", self.version)


class CapabilitiesRenderParams(object):
    def __init__(self, coverages, version, sections=None, accept_languages=None,
                 accept_formats=None, updatesequence=None, request=None):
        self._coverages = coverages
        self._version = version
        self._sections = sections or ()
        self._accept_languages = accept_languages or ()
        self._accept_formats = accept_formats or ()
        self._updatesequence = updatesequence
        self._request = request

    coverages        = property(lambda self: self._coverages)
    version          = property(lambda self: self._version)
    sections         = property(lambda self: self._sections)
    accept_languages = property(lambda self: self._accept_languages)
    accept_formats   = property(lambda self: self._accept_formats)
    updatesequence   = property(lambda self: self._updatesequence)

    @property
    def request(self):
        return self._request

    @request.setter
    def request(self, value):
        self._request = value

    def __iter__(self):
        yield "request", "GetCapabilities"

        if self.sections:
            yield "sections", ",".join(self.sections)

        if self.accept_languages:
            yield "acceptlanguages", ",".join(self.accept_languages)

        if self.accept_formats:
            yield "acceptformats", ",".join(self.accept_formats)

        if self.updatesequence:
            yield "updatesequence", self.updatesequence



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

"""
This module contains utils for XML encoding, decoding and printing.
"""
from django.utils.six import string_types
import logging
try:
    from lxml import etree
except ImportError:
    try:
        # Python 2.5
        import xml.etree.cElementTree as etree
    except ImportError:
        try:
            # Python 2.5
            import xml.etree.ElementTree as etree
        except ImportError:
            try:
                # normal cElementTree install
                import cElementTree as etree
            except ImportError:
                try:
                    # normal ElementTree install
                    import elementtree.ElementTree as etree
                except ImportError:
                    pass  # could not get etree


logger = logging.getLogger(__name__)


#-----------------------------------------------------------------------
# lxml.etree helpers
#-----------------------------------------------------------------------


class NameSpace(object):
    """ Helper object to ease the dealing with namespaces in both encoding and
        decoding.

        :param uri: the namespace URI
        :param prefix: the namespace prefix
        :param schema_location: the schema location of this namespace
    """

    def __init__(self, uri, prefix=None, schema_location=None):
        self._uri = uri
        self._lxml_uri = "{%s}" % uri
        self._prefix = prefix
        self._schema_location = schema_location

    @property
    def uri(self):
        return self._uri

    @property
    def prefix(self):
        return self._prefix

    @property
    def schema_location(self):
        return self._schema_location

    def __eq__(self, other):
        if isinstance(other, NameSpace):
            return self.uri == other.uri
        elif isinstance(other, string_types):
            return self.uri == other

        raise TypeError

    def __call__(self, tag):
        return self._lxml_uri + tag


class NameSpaceMap(dict):
    """ Helper object to ease the setup and management of namespace collections
        in both encoding and decoding. Can (and should) be passed as
        ``namespaces`` attribute in :class:`eoxserver.core.decoders.xml.Decoder`
        subclasses.

        :param namespaces: a list of :class:`NameSpace` objects.
    """

    def __init__(self, *namespaces):
        self._schema_location_dict = {}
        for namespace in namespaces:
            self.add(namespace)
        self._namespaces = namespaces

    def add(self, namespace):
        self[namespace.prefix] = namespace.uri
        if namespace.schema_location:
            self._schema_location_dict[namespace.uri] = (
                namespace.schema_location
            )

    def __copy__(self):
        return type(self)(*self._namespaces)

    @property
    def schema_locations(self):
        return self._schema_location_dict


def parse(obj):
    """ Helper function to parse XML either directly from a string, or fall back
        to whatever ``lxml.etree.parse`` parses. Returns ``None`` if it could
        not parse any XML.
    """
    if etree.iselement(obj):
        return obj.getroottree()
    elif isinstance(obj, etree._ElementTree):
        return obj
    try:
        tree_or_elem = etree.fromstring(obj.encode('ascii'))
        if etree.iselement(tree_or_elem):
            return tree_or_elem.getroottree()
        return tree_or_elem
    except:
        pass
    try:
        return etree.parse(obj)
    except:
        pass


def add_cdata(element, cdata):
    assert not element.text, \
        "Can't add a CDATA section. Element already has some text: %r" % (
            element.text
        )

    element.text = cdata

try:
    typemap = {
        etree.CDATA: add_cdata
    }
except AttributeError:
    typemap = {}

ns_xsi = NameSpace("http://www.w3.org/2001/XMLSchema-instance", "xsi")


class XMLEncoder(object):
    """ Base class for XML encoders using lxml.etree. This class does not
        actually provide any helpers for encoding XML in a tree structure (this
        is already done in lxml.etree), but adds tree to string serialization
        and automatic handling of schema locations.
    """

    def serialize(self, tree, pretty_print=True, encoding='iso-8859-1'):
        """ Serialize a tree to an XML string. Also adds the ``schemaLocations``
            attribute to the root node.
        """
        schema_locations = self.get_schema_locations()
        # Both Element and ElementTree have to be supported.
        root = tree if hasattr(tree, 'attrib') else tree.getroot()
        root.attrib[ns_xsi("schemaLocation")] = " ".join(
            "%s %s" % (uri, loc) for uri, loc in schema_locations.items()
        )

        return etree.tostring(
            tree, pretty_print=pretty_print, encoding=encoding
        )

    @property
    def content_type(self):
        return "text/xml"

    def get_schema_locations(self):
        """ Interface method. Returns a dict mapping namespace URIs to a network
            locations.
        """
        return {}



import os.path
from zipfile import ZipFile
import json
try:
    from io import StringIO
except ImportError:
    from cStringIO import StringIO
import traceback
import re
import mimetypes
import shutil

from django.http import (
    HttpResponse, HttpResponseBadRequest, HttpResponseNotAllowed, FileResponse
)
from django.contrib.gis.geos import GEOSGeometry
from django.db import transaction
from django.db.models import Q
from django.shortcuts import get_object_or_404

from eoxserver.core.config import get_eoxserver_config
from eoxserver.core.decoders import config
from eoxserver.backends.access import vsi_open
from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.registration.product import (
    ProductRegistrator
)
from eoxserver.resources.coverages.registration.browse import BrowseRegistrator
from eoxserver.resources.coverages.registration.registrators.gdal import (
    GDALRegistrator
)

# def browse_view(request, identifier):
#     browse_type = request.GET.get('type')
#     style = request.GET.get('style')

#     qs = models.Browse.objects.filter(
#         product__identifier=identifier,
#         style=style
#     )

#     if browse_type:
#         qs = qs.filter(browse_type__name=browse_type)
#     else:
#         qs = qs.filter(browse_type__isnull=True)

#     browse = qs.get()

#     ds = gdal.Open(get_vsi_path(browse))
#     tmp_file = vsi.TemporaryVSIFile.from_buffer('')
#     driver = gdal.GetDriverByName('PNG')

#     gt = ds.GetGeoTransform()

#     out_ds = driver.Create(tmp_file.name, 500, 500, 3)
#     out_ds.SetGeoTransform([
#         gt[0],
#     )


#     driver.CreateCopy(tmp_file.name, ds)

#     ds = None

#     return HttpResponse(tmp_file.read(), content_type='image/png')


def metadata(request, identifier, semantic):
    """ View to retrieve metadata files for a specific product.
    """
    if request.method != 'GET':
        return HttpResponseNotAllowed(['GET'])

    frmt = request.GET.get('format')

    semantic_code = {
        name: code
        for code, name in models.MetaDataItem.SEMANTIC_CHOICES
    }[semantic]

    qs = models.MetaDataItem.objects.filter(
        eo_object__identifier=identifier, semantic=semantic_code,
    )
    if frmt:
        qs = qs.filter(format=frmt)

    metadata_item = get_object_or_404(qs)

    return FileResponse(
        vsi_open(metadata_item), content_type=metadata_item.format
    )


def product_register(request):
    """ View to register a Product + 'Granules' (coverages) from a so-called
        'product.zip', entailing metadata and referencing local files.
    """
    if request.method != 'POST':
        return HttpResponseNotAllowed(['POST'])

    content = request.read()

    try:
        buffered_file = StringIO(content)
        zipfile = ZipFile(buffered_file)
    except Exception as e:
        return HttpResponseBadRequest('Failed to open ZIP file: %s' % e)

    try:
        with zipfile as zipfile, transaction.atomic():
            product_desc = json.load(zipfile.open('product.json'))
            granules_desc = json.load(zipfile.open('granules.json'))

            # get the collection from the 'parentId'
            try:
                parent_id = product_desc['properties']['eop:parentIdentifier']
                collection = models.Collection.objects.get(identifier=parent_id)
            except KeyError:
                return HttpResponseBadRequest(
                    'Missing product property: eop:parentIdentifier'
                )
            except models.Collection.DoesNotExist:
                return HttpResponseBadRequest(
                    'No such collection %r' % parent_id
                )

            product = _register_product(collection, product_desc, granules_desc)

            _add_metadata(
                product, zipfile, 'description.html', 'documentation',
                'text/html',
            )
            _add_metadata(
                product, zipfile, 'thumbnail\.(png|jpeg|jpg)', 'thumbnail'
            )
            _add_metadata(
                product, zipfile, 'metadata\.xml', 'description', 'text/xml'
            )

            granules = []
            # iterate over the granules and register them
            for granule_desc in granules_desc['features']:
                coverage = _register_granule(
                    product, collection, granule_desc
                )
                granules.append(coverage)

                # add the coverage to the product
                models.product_add_coverage(product, coverage)

            models.collection_insert_eo_object(collection, product)
            models.collection_collect_metadata(
                collection, product_summary=True, coverage_summary=True
            )

    except (KeyError, ValueError) as e:
        return HttpResponseBadRequest(str(e))
    except Exception:
        return HttpResponseBadRequest(traceback.format_exc())

    return HttpResponse(
        'Successfully registered product %s with granules: %s'
        % (product.identifier, ', '.join(
            granule.identifier for granule in granules
        ))
    )


def _register_product(collection, product_def, granules_def):
    type_name = None
    collection_type = collection.collection_type

    # get the first product type from the collection
    if collection_type:
        product_type = collection_type.allowed_product_types.first()
        if product_type:
            type_name = product_type.name

    properties = product_def['properties']

    footprint = GEOSGeometry(json.dumps(product_def['geometry'])).wkt
    identifier = properties['eop:identifier']
    begin_time = properties['timeStart']
    end_time = properties['timeEnd']

    location = properties['originalPackageLocation']

    product, _ = ProductRegistrator().register(
        metadata_locations=[],
        mask_locations=[],
        package_path=location,
        overrides=dict(
            identifier=identifier,
            footprint=footprint,
            begin_time=begin_time,
            end_time=end_time,
            **properties
        ),
        type_name=type_name,
        discover_masks=False,
        discover_browses=False,
        discover_metadata=False,
        replace=True,
    )

    browse_locations = []
    features = granules_def['features']
    if len(features) == 1:
        location = features[0]['properties'].get('location')
        if location:
            browse_locations.append(location)
    else:
        browse_locations = [
            granule_desc['properties']['location']
            for granule_desc in features
            if granule_desc['properties'].get('band') == 'TCI'
        ]
    for browse_location in browse_locations:
        BrowseRegistrator().register(
            product.identifier, [browse_location]
        )

    return product


def _register_granule(product, collection, granule_def):
    properties = granule_def['properties']
    coverage_types_base = models.CoverageType.objects.filter(
        allowed_collection_types__collections=collection
    )

    if 'band' in properties:
        # get the coverage type associated with the collection and the granules
        # band ID
        identifier = '%s_%s' % (product.identifier, properties['band'])
        coverage_type = coverage_types_base.get(
            name__iendswith=properties['band']
        )

    else:
        # for a lack of a better generic way, just get the first allowed
        # coverage type associated with the collection
        identifier = os.path.basename(properties['location'])
        coverage_type = coverage_types_base[0]

    overrides = dict(
        identifier=identifier,
        begin_time=product.begin_time,
        end_time=product.end_time,
        footprint=GEOSGeometry(json.dumps(granule_def['geometry'])).wkt
    )

    return GDALRegistrator().register(
        data_locations=[[properties['location']]],
        metadata_locations=[],
        coverage_type_name=coverage_type.name,
        overrides=overrides,
        replace=True,
    ).coverage


def _add_metadata(product, zipfile, pattern, semantic, frmt=None):
    def _get_file_info(zipfile, pattern):
        for info in zipfile.infolist():
            if re.match(pattern, info.filename):
                return info

    reader = RegistrationConfigReader(get_eoxserver_config())
    metadata_filename_template = reader.metadata_filename_template

    info = _get_file_info(zipfile, pattern)
    if info and metadata_filename_template:
        frmt = frmt or mimetypes.guess_type(info.filename)[0]

        semantic_code = {
            name: code
            for code, name in models.MetaDataItem.SEMANTIC_CHOICES
        }[semantic]

        out_filename = metadata_filename_template.format(
            product_id=product.identifier, filename=info.filename
        )

        out_dirname = os.path.dirname(out_filename)

        # make directories
        try:
            os.makedirs(out_dirname)
        except OSError as exc:
            if exc.errno != 17:
                raise

        with open(out_filename, "w") as out_file:
            shutil.copyfileobj(zipfile.open(info), out_file)

        models.MetaDataItem.objects.create(
            eo_object=product, format=frmt, location=out_filename,
            semantic=semantic_code
        )


class RegistrationConfigReader(config.Reader):
    section = "coverages.registration"
    metadata_filename_template = config.Option()



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


class LayerNotDefined(Exception):
    def __init__(self, layer):
        super(LayerNotDefined, self).__init__("No such layer '%s'." % layer)

    locator = "layers"
    code = "LayerNotDefined"


class InvalidCRS(Exception):
    def __init__(self, value, crs_param_name):
        super(InvalidCRS, self).__init__(
            "Invalid '%s' parameter value: '%s'"
            % (crs_param_name.upper(), value)
        )
        self.locator = crs_param_name
    code = "InvalidCRS"


class InvalidFormat(Exception):
    def __init__(self, value):
        super(InvalidFormat, self).__init__(
            "Unknown format name '%s'" % value
        )
    locator = "format"
    code = "InvalidFormat"



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import logging
from unittest import SkipTest

try:
    from twisted.protocols.ftp import FTPFactory, FTPRealm
    from twisted.cred.portal import Portal
    from twisted.cred.checkers import AllowAnonymousAccess, FilePasswordDB
    from twisted.internet import reactor
    HAVE_TWISTED = True
except ImportError:
    HAVE_TWISTED = False



logger = logging.getLogger(__name__)

def withFTPServer(port=2121, directory=None):
    def wrap(f):
        def wrapped(*args, **kwargs):
            if not HAVE_TWISTED:
                raise SkipTest("This test requires Twisted to run.")
            # TODO: start FTP server
            ret = f(args, **kwargs)
            # TODO: stop ftp server
            return ret
        return wrapped
    return wrap



# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-11-13 15:38
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('coverages', '0002_browse_type_fields'),
    ]

    operations = [
        migrations.AddField(
            model_name='metadataitem',
            name='semantic',
            field=models.SmallIntegerField(blank=True, choices=[(0, b'other'), (1, b'description'), (2, b'documentation'), (3, b'thumbnail')], null=True),
        ),
        migrations.AlterUniqueTogether(
            name='metadataitem',
            unique_together=set([('eo_object', 'semantic')]),
        ),
    ]



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from django.contrib.gis.geos import Polygon


EXTENT_EPSG_4326 = Polygon.from_bbox((-180, -90, 180, 90))
EXTENT_EPSG_4326.srid = 4326

CRS_WIDTH = {
    3857: -20037508.3428 * 2,
    4326: 360.0
}


def extent_crosses_dateline(extent, srid=4326):
    """ Returns ``True`` if a dataset crosses the dateline border. """

    poly = Polygon.from_bbox(extent)
    poly.srid = srid

    poly.transform(EXTENT_EPSG_4326.srs)

    if not EXTENT_EPSG_4326.contains(poly):
        return True

    return False


def wrap_extent_around_dateline(extent, srid=4326):
    """ Wraps the given extent around the dateline. Currently only works for
    EPSG:4326 and EPSG:3857"""

    try:
        return (extent[0] - CRS_WIDTH[srid], extent[1],
                extent[2] - CRS_WIDTH[srid], extent[3])
    except KeyError:
        raise NotImplementedError(
            "Dateline wrapping is not implemented for SRID "
            "%d. Supported are SRIDs %s." %
            (srid, ", ".join(CRS_WIDTH.keys()))
        )



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import re

from django.core.management.base import CommandError, BaseCommand
from django.db import transaction
from django.contrib.gis.geos import GEOSGeometry

from eoxserver.core.util.timetools import parse_iso8601
from eoxserver.backends.storages import get_handler_class_for_model
from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.management.commands import (
    CommandOutputMixIn, SubParserMixIn
)
from eoxserver.resources.coverages.registration.product import (
    ProductRegistrator
)
from eoxserver.resources.coverages.registration.exceptions import (
    RegistrationError
)


class Command(CommandOutputMixIn, SubParserMixIn, BaseCommand):
    """ Command to manage product types. This command uses sub-commands for the
        specific tasks: register, deregister
    """
    def add_arguments(self, parser):
        register_parser = self.add_subparser(parser, 'register')
        deregister_parser = self.add_subparser(parser, 'deregister')
        discover_parser = self.add_subparser(parser, 'discover')

        register_parser.add_argument(
            '--identifier', '-i', default=None,
            help='Override the identifier of the to-be registered product.'
        )

        register_parser.add_argument(
            "--identifier-template",
            dest="identifier_template", default=None,
            help="Add a template to construct the final identifier."
        )

        register_parser.add_argument(
            '--footprint', default=None,
            help='Override the footprint of the to-be registered product.'
        )

        register_parser.add_argument(
            '--simplify-footprint',
            dest='simplify_footprint_tolerance', nargs='?',
            default=None, type=float,
            help=(
                'Simplify the footprint. Optionally specify a tolerance value.'
            )
        )

        register_parser.add_argument(
            '--begin-time', default=None, type=parse_iso8601,
            help='Override the begin time of the to-be registered product.'
        )

        register_parser.add_argument(
            '--end-time', default=None, type=parse_iso8601,
            help='Override the end time of the to-be registered product.'
        )

        register_parser.add_argument(
            '--set', '-s', dest='set_overrides',
            nargs=2, default=[], action='append',
            help=(
                'Set (or override) additional metadata tags like '
                '"opt:cloudCover".'
            )
        )

        register_parser.add_argument(
            '--metadata-file',
            dest='metadata_locations', nargs='+', default=[], action='append',
            help=(
                'Add metadata file to associate with the product. '
                'List of items. Can be specified multiple times.'
            )
        )

        register_parser.add_argument(
            '--type', '--product-type', '-t', dest='type_name', default=None,
            help=(
                'The name of the product type to associate the product with. '
                'Optional.'
            )
        )

        register_parser.add_argument(
            '--mask', '-m', dest='mask_locations', default=[], action='append',
            nargs=2,
            help=(
                'Add a mask to associate with the product. List of items, '
                'first one is the mask name, the rest is the location '
                'definition. Can be specified multiple times.'
            )
        )

        register_parser.add_argument(
            '--mask-geometry', '-g', dest='mask_geometries', default=[],
            action='append', nargs=2,
            help=(
                'Add a mask to associate with the product. List of items, '
                'first one is the mask name, second is the mask geometry in '
                'WKT. Can be specified multiple times.'
            )
        )

        register_parser.add_argument(
            '--no-extended-metadata', dest='extended_metadata',
            default=True, action='store_false',
            help=(
                'When this flag is set, only the basic metadata (identifier, '
                'footprint, begin- and end-time) is stored.'
            )
        )

        register_parser.add_argument(
            '--no-masks', dest='discover_masks',
            default=True, action='store_false',
            help=(
                'When this flag is set, no masks will be discovered.'
            )
        )

        register_parser.add_argument(
            '--no-browses', dest='discover_browses',
            default=True, action='store_false',
            help=(
                'When this flag is set, no browses will be discovered.'
            )
        )

        register_parser.add_argument(
            '--no-metadata', dest='discover_metadata',
            default=True, action='store_false',
            help=(
                'When this flag is set, no metadata will be discovered.'
            )
        )

        register_parser.add_argument(
            '--package', '-p', default=None,
            help=(
                'The path to a storage (directory, ZIP-file, etc.).'
            )
        )

        register_parser.add_argument(
            "--collection", "--collection-identifier", "-c",
            dest="collection_identifiers", action="append", default=[],
            help="Add the product to the specified collection."
        )

        register_parser.add_argument(
            "--replace", "-r",
            dest="replace", action="store_true", default=False,
            help=(
                "Optional. If the product with the given identifier already "
                "exists, replace it. Without this flag, this would result in "
                "an error."
            )
        )

        register_parser.add_argument(
            '--print-identifier', dest='print_identifier',
            default=False, action='store_true',
            help=(
                'When this flag is set, only the identifier of the registered '
                'product will be printed to stdout.'
            )
        )
        deregister_parser.add_argument(
            '--all', '-a', action="store_true",
            default=False, dest='all_products',
            help='When this flag is set, all the products are selected to be derigesterd'
        )


        deregister_parser.add_argument(
                'identifier', default=None, nargs='?',
                help='The identifier of the product to deregister.'
            )

        discover_parser.add_argument(
                'identifier', default=None,
                help='The identifier of the product to descover.'
            )

        discover_parser.add_argument(
            'pattern', nargs='?', default=None,
            help='A glob path pattern to limit the search.'
        )

        # TODO: only via 'browse' command?
        # register_parser.add_argument(
        #     '--browse', '-b',
        #     dest='browse_handles', default=None, action='append',
        #     # help='The name of the grid to associate the product with.'
        # )

    @transaction.atomic
    def handle(self, subcommand, *args, **kwargs):
        """ Dispatch sub-commands: register, deregister.
        """
        if subcommand == "register":
            self.handle_register(*args, **kwargs)
        elif subcommand == "deregister":
            self.handle_deregister(*args, **kwargs)
        elif subcommand == "discover":
            self.handle_discover(kwargs.pop('identifier')[0], *args, **kwargs)

    def handle_register(self, **kwargs):
        """ Handle the creation of a new product
        """
        try:
            overrides = dict(
                identifier=kwargs['identifier'],
                footprint=kwargs['footprint'],
                begin_time=kwargs['begin_time'],
                end_time=kwargs['end_time'],
            )

            for name, value in kwargs['set_overrides']:
                overrides[convert_name(name)] = value

            mask_locations = kwargs['mask_locations'] + [
                (name, GEOSGeometry(geom))
                for name, geom in kwargs['mask_geometries']
            ]

            product, replaced = ProductRegistrator().register(
                metadata_locations=kwargs['metadata_locations'],
                mask_locations=mask_locations,
                package_path=kwargs['package'],
                overrides=overrides,
                identifier_template=kwargs['identifier_template'],
                type_name=kwargs['type_name'],
                extended_metadata=kwargs['extended_metadata'],
                discover_masks=kwargs['discover_masks'],
                discover_browses=kwargs['discover_browses'],
                discover_metadata=kwargs['discover_metadata'],
                replace=kwargs['replace'],
                simplify_footprint_tolerance=kwargs.get(
                    'simplify_footprint_tolerance'
                ),
            )

            for collection_identifier in kwargs['collection_identifiers']:
                try:
                    collection = models.Collection.objects.get(
                        identifier=collection_identifier
                    )
                except models.Collection.DoesNotExist:
                    raise CommandError(
                        'No such collection %r' % collection_identifier
                    )
                models.collection_insert_eo_object(collection, product)

        except RegistrationError as e:
            raise CommandError('Failed to register product. Error was %s' % e)

        if kwargs['print_identifier']:
            print(product.identifier)
        else:
            self.print_msg(
                'Successfully registered product %r' % product.identifier
            )

    def handle_deregister(self, identifier, all_products, *args, **kwargs):
        """ Handle the deregistration a product
        """
        if not all_products and not identifier:
            raise CommandError('please specify a product/s to remove')
        else:
            if all_products:
                products = models.Product.objects.all()
            elif identifier:
                products = [models.Product.objects.get(identifier=identifier)]
            for product in products:
                try:
                    product_id = product.identifier
                    product.delete()
                    self.print_msg('Successfully deregistered product %r' % product_id)
                except models.Product.DoesNotExist:
                    raise CommandError('No such Product %r' % identifier)

    def handle_discover(self, identifier, pattern, *args, **kwargs):
        try:
            product = models.Product.objects.get(identifier=identifier)
        except models.Product.DoesNotExist:
            raise CommandError('No such Product %r' % identifier)

        package = product.package
        if package:
            handler_cls = get_handler_class_for_model(package)
            if handler_cls:
                with handler_cls(package.url) as handler:
                    for item in handler.list_files(pattern):
                        print(item)


def camel_to_underscore(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def convert_name(name):
    namespace, _, sub_name = name.partition(':')
    if namespace in ('eop', 'opt', 'sar', 'alt'):
        return camel_to_underscore(sub_name)
    return camel_to_underscore(name)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import os
import os.path
import shutil
import tempfile
import errno
import logging
import threading

from eoxserver.core.config import get_eoxserver_config
from eoxserver.backends.config import CacheConfigReader


logger = logging.getLogger(__name__)


# global instance of the cache context
cache_context_storage = threading.local()


class CacheException(Exception):
    pass


def setup_cache_session(config=None):
    """ Initialize the cache context for this session. If a cache context was
        already present, an exception is raised.
    """
    if not config:
        config = CacheConfigReader(get_eoxserver_config())

    set_cache_context(
        CacheContext(config.retention_time, config.directory, True)
    )


def shutdown_cache_session():
    """ Shutdown the cache context for this session and trigger any pending
        cleanup actions required.
    """
    try:
        cache_context = get_cache_context()
        cache_context.cleanup()
    except CacheException:
        # it seems that the cache was already shut down.
        pass
    set_cache_context(None)


def set_cache_context(cache_context):
    """ Sets the cache context for this session. Raises an exception if there
        was already a cache context associated.
    """
    if cache_context is not None:
        if getattr(cache_context_storage, "cache_context", None) is not None:
            raise CacheException(
                "The cache context for this session was already initialized."
            )

    cache_context_storage.cache_context = cache_context


def get_cache_context():
    """ Get the thread local cache context for this session. Raises an exception
        if the session was not initialized.
    """
    cache_context = getattr(cache_context_storage, "cache_context", None)
    if not cache_context:
        raise CacheException(
            "The cache context for this session was not initialized."
        )
    return cache_context


class CacheContext(object):
    """ Context manager to manage cached files.
    """
    def __init__(self, retention_time=None, cache_directory=None, managed=False):
        self._cached_objects = set()

        if not cache_directory:
            cache_directory = tempfile.mkdtemp(prefix="eoxs_cache")
            self._temporary_dir = True
        else:
            self._temporary_dir = False

        self._cache_directory = cache_directory
        self._retention_time = retention_time
        self._level = 0
        self._mappings = {}

        self._managed = managed

    @property
    def cache_directory(self):
        """ Returns the configured cache directory.
        """
        return self._cache_directory

    def relative_path(self, cache_path):
        """ Returns a path relative to the cache directory.
        """
        return os.path.join(self._cache_directory, cache_path)

    def add_mapping(self, path, item):
        """ Add an external file to this context. Those files will be treated as
            if they are "within" the caches directory, but will not be clead up
            afterwards.
        """
        self._mappings[path] = item

    def add_path(self, cache_path):
        """ Add a path to this cache context. Also creates necessary
            sub-directories.
        """
        self._cached_objects.add(cache_path)
        relative_path = self.relative_path(cache_path)

        try:
            # create all necessary subdirectories
            os.makedirs(os.path.dirname(relative_path))
        except OSError as e:
            # it's only ok if the dir already existed
            if e.errno != errno.EEXIST:
                raise

        return relative_path

    def cleanup(self):
        """ Perform cache cleanup.
        """
        if self._retention_time and not self._temporary_dir:
            # no cleanup required
            return

        elif not self._temporary_dir:
            for path in self._cached_objects:
                os.remove(path)
            self._cached_objects.clear()

        else:
            shutil.rmtree(self._cache_directory)
            self._cached_objects.clear()

    def contains(self, cache_path):
        """ Check whether or not the path is contained in this cache.
        """
        if cache_path in self._cached_objects:
            return True

        return os.path.exists(self.relative_path(cache_path))

    def __contains__(self, cache_path):
        """ Alias for method `contains`.
        """
        return self.contains(cache_path)

    def __enter__(self):
        """ Context manager protocol, for recursive use. Each time the a context
            is entered, the internal level is raised by one.
        """
        self._level += 1
        return self

    def __exit__(self, etype=None, evalue=None, tb=None):
        """ Exit of context manager protocol. Performs cache cleanup if
            the level drops to zero.
        """
        self._level -= 1
        if self._level == 0 and not self._managed:
            self.cleanup()



#!/usr/bin/python
#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import sys
import re
import argparse
import traceback
import textwrap
from os.path import splitext
import logging

from eoxserver.core.util.timetools import getDateTime
from eoxserver.core.util.xmltools import DOMElementToXML
from eoxserver.processing.preprocessing.util import  check_file_existence
from eoxserver.processing.preprocessing import (
    WMSPreProcessor, RGBA, ORIG_BANDS, NativeMetadataFormatEncoder
)
from eoxserver.processing.preprocessing.format import (
    get_format_selection, GeoTIFFFormatSelection
)
from eoxserver.processing.preprocessing.georeference import (
    Extent, GCPList
)


def main(args):
    # create a parser instance
    parser = argparse.ArgumentParser(add_help=True, fromfile_prefix_chars='@',
                                     argument_default=argparse.SUPPRESS,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    
    parser.description = textwrap.dedent("""\
    Takes <infile> raster data and pre-processes it into 
    <outfiles_basename>.tif, a GeoTIFF converted to RGB using default internal 
    tiling, internal overviews, no compression, and 0 as no-data value, and 
    <outfiles_basename>.xml, a EOxServer simple XML EO metadata file.

    The outfiles are ready to be used with the eoxs_register command of an 
    EOxServer instance.

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!Caution, you will loose data quality!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    The data quality of <outfiles_basename>.tif is most likely different from 
    <infile> because of the pre-processing applied. Reasons include 
    re-projection, compression, conversion to RGB, indexing, etc.
    
    Examples:
    
    # basic usage with no creation of metadata
    eoxserver-preprocess.py --no-metadata input.tif
    
    # basic usage with creation of metadata with specific basename
    eoxserver-preprocess.py --coverage-id=a --begin-time=2008-03-01T13:00:00Z \\ 
                            --end-time=2008-03-01T13:00:00Z input.tif output
                            
    # with RGBA band selection
    eoxserver-preprocess.py --bands 1:0:255,2:0:255,3:0:100,4 --rgba \\
                            --no-metadata input.tif

    # with DEFLATE compression and color index from a palette file
    eoxserver-preprocess.py  --compression=DEFLATE --zlevel=2 --indexed \\ 
                             --pct palette.vrt --no-metadata input.tif
    
    # (re-)setting the extent of the file
    eoxserver-preprocess.py --extent 0,0,10,10,4326 --no-metadata input.tif
    
    # Using GCPs
    eoxserver-preprocess.py --no-metadata --gcp 0,0,10,10 --gcp 2560,0,50,10 \\
                            --gcp 0,1920,10,50 --gcp 2560,1920,50,50 \\ 
                            --georef-crs 4326 input.tif
                            
    # reading arguments from a file (1 line per argument), with overrides
    eoxserver-preprocess.py @args.txt --crs=3035 --no-tiling input.tif
    """)
    
    #===========================================================================
    # Metadata parsing group
    #===========================================================================
    
    # TODO: won't work with mutual exclusive groups. Bug in Argparse?
    #md_g = parser.add_mutually_exclusive_group(required=True)
    
    #md_g.add_argument("--no-metadata", dest="generate_metadata",
    #                  action="store_false",
    #                  help="Explicitly turn off the creation of a metadata " 
    #                       "file.")
    
    #md_g_data = md_g.add_argument_group()
    
    #md_g_data.add_argument("--begin-time", dest="begin_time", 
    #                       type=_parse_datetime,
    #                       help="The ISO 8601 timestamp of the begin time.")
    #md_g_data.add_argument("--end-time", dest="end_time",
    #                       type=_parse_datetime,
    #                       help="The ISO 8601 timestamp of the end time.")
    #md_g_data.add_argument("--coverage-id", dest="coverage_id",
    #                       type=_parse_coverage_id, 
    #                       help="The ID of the coverage, must be a valid "
    #                            "NCName.")
    
    # should be mutually exclusive
    parser.add_argument("--no-metadata", dest="generate_metadata",
                        action="store_false",
                        help="Explicitly turn off the creation of a metadata " 
                             "file.")
    parser.add_argument("--begin-time", dest="begin_time", type=_parse_datetime,
                        help="The ISO 8601 timestamp of the begin time.")
    parser.add_argument("--end-time", dest="end_time", type=_parse_datetime,
                        help="The ISO 8601 timestamp of the end time.")
    parser.add_argument("--coverage-id", dest="coverage_id",
                        type=_parse_coverage_id, 
                        help="The ID of the coverage, must be a valid NCName.")
    
    #===========================================================================
    # Georeference group
    #===========================================================================
    
    georef_g = parser.add_mutually_exclusive_group()
    georef_g.add_argument("--extent", dest="extent", type=_parse_extent,
                          help="The extent of the dataset, as a 4-tuple of "
                               "floats.")
    #georef_g.add_argument("--footprint", dest="footprint", 
    #                      type=_parse_footprint,
    #                      help="The footprint of the dataset, as a Polygon WKT.")
    georef_g.add_argument("--gcp", dest="gcps", type=_parse_gcp,
                          action="append",
                          help="A Ground Control Point in the format: "
                               "'pixel,line,easting,northing[,elevation]'.")
    
    parser.add_argument("--georef-crs", dest="georef_crs")
    
    #===========================================================================
    # Arbitraries
    #===========================================================================
    
    parser.add_argument("--crs", dest="crs", type=int, 
                        help="The desired output CRS ID of the dataset.")
    parser.add_argument("--no-tiling", dest="tiling", action="store_false",
                        default=True, 
                        help="Explicitly turn of the tiling of the output "
                             "dataset.")
    parser.add_argument("--no-overviews", dest="overviews",
                        action="store_false", default=True,
                        help="Explicitly turn of the creation of overview "
                             "images of the output dataset.")
    
    #===========================================================================
    # Bands group
    #===========================================================================
    
    bands_g = parser.add_mutually_exclusive_group()
    
    bands_g.add_argument("--rgba", dest="bandmode", action="store_const",
                         const=RGBA,
                         help="Convert the image to RGBA, using the first four "
                              "bands.")
    bands_g.add_argument("--orig-bands", dest="bandmode", action="store_const",
                         const=ORIG_BANDS,
                         help="Explicitly keep all original bands.")
    
    parser.add_argument("--bands", dest="bands", type=_parse_bands, 
                        help="A comma separated list of bands with optional "
                             "subsets in the form: 'no[:low:high]'. Either "
                             "three bands, or four when --rgba is requested.")

    parser.add_argument("--footprint-alpha", dest="footprint_alpha",
                        action="store_true", default=False,
                        help="A comma separated list of bands with optional "
                             "subsets in the form: 'no[:low:high]'. Either "
                             "three bands, or four when --rgba is requested.")
    
    parser.add_argument("--compression", dest="compression",
                        choices=GeoTIFFFormatSelection.SUPPORTED_COMPRESSIONS,
                        help="The desired compression technique.")
    parser.add_argument("--jpeg-quality", dest="jpeg_quality", type=int,
                        help="The JPEG algorithm quality when JPEG compression "
                             "is requested. Default is 75.")
    parser.add_argument("--zlevel", dest="zlevel", type=int,
                        help="The zlevel quality setting when DEFLATE "
                             "compression is requested.")
    
    indexed_group = parser.add_argument_group()
    
    # TODO: pct depends on indexed
    
    parser.add_argument("--indexed", dest="color_index", action="store_true",
                        help="Create a paletted (indexed) image.")
    parser.add_argument("--pct", dest="palette_file",
                        help="Use the given file as color palette. If not "
                             "given, a new palette will be calculated.")
    
    parser.add_argument("--no-data-value", dest="no_data_value", 
                        type=_parse_nodata_values,
                        help="Either one, or a list of no-data values.")
    
    parser.add_argument("--co", dest="creation_options", type=int, 
                        action="append",
                        help="Additional GDAL dataset creation options. "
                             "See http://www.gdal.org/frmt_gtiff.html")
    
    parser.add_argument("--traceback", action="store_true", default=False)
    
    parser.add_argument("--force", "-f", dest="force", action="store_true",
                        default=False,
                        help="Override files, if they already exist.")

    parser.add_argument("--verbosity", "-v", dest="verbosity", type=int,
                        default=1,
                        help="Set the verbosity (0, 1, 2). Default is 1.")
    
    parser.add_argument("input_filename", metavar="infile", nargs=1,
                        help="The input raster file to be processed.")
    parser.add_argument("output_basename", metavar="outfiles_basename",
                        nargs="?", 
                        help="The base name of the outputfile(s) to be "
                             "generated.")
    
    values = vars(parser.parse_args(args))
    
    
    # check metadata values
    if "generate_metadata" in values and ("begin_time" in values 
                                          or "end_time" in values
                                          or "coverage_id" in values):
        parser.error("--no-metadata is mutually exclusive with --begin-time, "
                     "--end-time and --coverage-id.")

    elif "generate_metadata" not in values and not ("begin_time" in values 
                                                    and "end_time" in values
                                                    and "coverage_id" in values):
        parser.error("Enter the full metadata with --begin-time, --end-time "
                     "and --coverage-id.")
    
    # hack to flatten the list
    values["input_filename"] = values["input_filename"][0]
    
    georef_crs = values.pop("georef_crs", None)
    
    if "extent" in values:
        values["geo_reference"] = Extent(*values.pop("extent"), srid=georef_crs or 4326)
    
    if "gcps" in values:
        values["geo_reference"] = GCPList(values.pop("gcps"), georef_crs or 4326)
    
    if "palette_file" in values and not "color_index" in values:
        parser.error("--pct can only be used with --indexed")
    
    # Extract format and execution specific values
    format_values = _extract(values, ("tiling", "compression", "jpeg_quality", 
                                      "zlevel", "creation_options"))
    exec_values = _extract(values, ("input_filename", "geo_reference",
                                    "generate_metadata"))
    other_values = _extract(values, ("traceback", ))
    metadata_values = _extract(values, ("coverage_id", "begin_time",
                                         "end_time"))

    force = values.pop("force", True)
    verbosity = values.pop("verbosity")
    output_basename = values.pop("output_basename", None)
    input_filename = exec_values.get("input_filename")

    # setup logging
    if verbosity > 0:
        if verbosity == 1: level = logging.WARN
        elif verbosity == 2: level = logging.INFO
        elif verbosity >= 3: level = logging.DEBUG
        logging.basicConfig(format="%(levelname)s: %(message)s", stream=sys.stderr,
                            level=level)
        

    try:
        # create a format selection
        format_selection = get_format_selection("GTiff", **format_values)


        # TODO: make 'tif' dependant on format selection
        # check files exist
        if not output_basename:
            output_filename = splitext(input_filename)[0] + "_proc" + format_selection.extension
            output_md_filename = splitext(input_filename)[0] + "_proc.xml"
        
        else:
            output_filename = output_basename + format_selection.extension
            output_md_filename = output_basename + ".xml"

        if not force:
            check_file_existence(output_filename)

        exec_values["output_filename"] = output_filename

        # create and run the preprocessor
        preprocessor = WMSPreProcessor(format_selection, **values)
        result = preprocessor.process(**exec_values)

        if exec_values.get("generate_metadata", True):
            if not force:
                check_file_existence(output_md_filename)

            encoder = NativeMetadataFormatEncoder()
            xml = DOMElementToXML(encoder.encodeMetadata(metadata_values["coverage_id"],
                                                         metadata_values["begin_time"],
                                                         metadata_values["end_time"],
                                                         result.footprint_raw))
            
            with open(output_md_filename, "w+") as f:
                f.write(xml)
        
    except Exception as e:
        # error wrapping
        if other_values["traceback"]:
            traceback.print_exc()
        sys.stderr.write("%s: %s\n" % (type(e).__name__, str(e)))


def _parse_datetime(input_str):
    """ Helper callback function to check if a given datetime is correct.
    """
    
    try:
        getDateTime(input_str)
    except:
        raise argparse.ArgumentTypeError("Wrong datetime format. Use ISO 8601.")
        
    return input_str
    


def _parse_coverage_id(input_str):
    """ Helper callback function to check if a given coverage ID is a valid 
        NCName.
    """
    
    if not re.match(r'^[a-zA-z_][a-zA-Z0-9_.-]*$', input_str):
        raise argparse.ArgumentTypeError("The given coverage ID '%s' is not a "
                                         "valid NCName." % input_str)
    
    return input_str


def _parse_extent(input_str):
    """ Helper callback function to parse an extent.
    """
    
    parts = input_str.split(",")
    if len(parts) != 4:
        raise argparse.ArgumentTypeError("Wrong format of extent.")
    
    return map(float, parts)


def _parse_footprint(input_str):
    """ Helper callback function to parse a footprint.
    """
    # TODO: implement
    pass


def _parse_gcp(input_str):
    """ Helper callback function to parse one GCP in the form 
        "pixel,line,easting,northing[,elevation]" and transform it into a list 
        like [x,y[,z],pixel,line]
    """
    
    parts = input_str.split(",")
    
    if not (len(parts) != 4 or len(parts) != 5):
        raise argparse.ArgumentTypeError("Wrong number of arguments for GCP.")
    
    try:
        gcps = [float(part) for part in parts]
        return gcps[2:] + gcps[:2]
    except ValueError:
        raise argparse.ArgumentTypeError("Wrong format of GCP.")


def _parse_bands(input_str):
    """ Helper callback function to parse a list of selected bands given from 
        the user.
    """
    
    result = []
    
    def parse_minmax(subparts, index):
        try:
            if subparts[index] in ("min", "max"):
                return subparts[index]
            else:
                return float(subparts[index])
        except IndexError:
            return None
        except ValueError:
            raise argparse.ArgumentTypeError("Wrong format of band subset.")
    
    for part in input_str.split(","):
        subparts = part.split(":")
        
        if len(subparts) < 1 or len(subparts) > 3:
            raise argparse.ArgumentTypeError("Wrong format of band.")
        
        number = int(subparts[0])
        dmin = parse_minmax(subparts, 1)
        dmax = parse_minmax(subparts, 2)
        result.append((number, dmin, dmax))
    
    return result


def _parse_nodata_values(input_str):
    """ Helper callback function to parse no-data values. 
    """
    
    return [int(part) for part in input_str.split(",")]


def _parse_creation_options(input_str):
    """ Helper callback function to parse additional dataset creation options. 
    """
    
    return input_str.split("=")


def _extract(dct, keys):
    """ Helper function to extract a list of keys from a dict and return these
        in a new dict.
    """
    tmp = {}
    for key in keys:
        try:
            tmp[key] = dct.pop(key)
        except KeyError:
            pass
    return tmp


if __name__ == "__main__":
    main(sys.argv[1:])



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


class MetadataReaderInterface(object):
    """ Interface for metadata readers.
    """

    def test(self, obj):
        """ Return a boolean value, whether or not metadata can be extracted 
            from the given object.
        """
        pass


    def format(self, obj):
        """ Returns a format specifier for the given object. Can be ignored, 
            when the reader only supports one format.
        """
        pass


    def read(self, obj):
        """ Returns a dict with any of the following keys:
            - identifier (string)
            - extent (a four tuple of floats)
            - size (a two-tuple of ints)
            - projection (an integer or two-tuple of two strings (definition and format))
            - footprint (a django.contrib.gis.geos.MultiPolygon)
            - begin_time (a python datetime.datetime)
            - end_time (a python datetime.datetime)

            The argument obj is of an arbitrary type, the reader needs to 
            determine whether or not the type is supported and an exception 
            shall be raised if not.
        """
        pass


class MetadataWriterInterface(object):
    """ Interface for metadata writers.
    """

    @property
    def formats(self):
        pass


    def write(self, values, file_obj, format=None):
        """ Write the given values (a dict) to the file-like object `file_obj`.
            The dict contains all of the following entries:
            - identifier (string)
            - extent (a four tuple of floats)
            - size (a two-tuple of ints)
            - projection (an integer or two-tuple of two strings (definition and format))
            - footprint (a django.contrib.gis.geos.MultiPolygon)
            - begin_time (a python datetime.datetime)
            - end_time (a python datetime.datetime)

            The writer may ignore non-applicable parameters.
        """
        pass


class GDALDatasetMetadataReaderInterface(object):
    """ Interface for GDAL dataset metadata readers.
    """

    def test_ds(self, obj):
        """ Return a boolean value, whether or not metadata can be extracted 
            from the given object.
        """
        pass


    def format(self, obj):
        """ Returns a format specifier for the given object. Can be ignored, 
            when the reader only supports one format.
        """
        pass


    def read_ds(self, ds):
        """ Returns a dict with any of the following keys:
            - identifier (string)
            - extent (a four tuple of floats)
            - size (a two-tuple of ints)
            - projection (an integer or two-tuple of two strings (definition and format))
            - footprint (a django.contrib.gis.geos.MultiPolygon)
            - begin_time (a python datetime.datetime)
            - end_time (a python datetime.datetime)

            The argument `ds` is a gdal.Dataset.
        """
        pass



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from eoxserver.backends.access import gdal_open
from eoxserver.resources.coverages.metadata.coverage_formats import (
    get_reader_by_test
)
from eoxserver.resources.coverages.registration.base import BaseRegistrator


class GDALRegistrator(BaseRegistrator):
    scheme = "GDAL"

    def _read_metadata_from_data(self, data_item, retrieved_metadata, cache, highest_resolution):
        ds = gdal_open(data_item)
        reader = get_reader_by_test(ds)
        if reader:
            values = reader.read(ds)

            format_ = values.pop("format", None)
            if format_:
                data_item.format = format_

            for key, value in values.items():
                if key == 'origin' and key in retrieved_metadata:
                    if value != retrieved_metadata[key]:
                        raise Exception('Different origin')

                if key == 'grid' and key in retrieved_metadata:
                    old = retrieved_metadata[key]
                    if isinstance(old, str):
                        continue

                    if old['axis_names'] != value['axis_names']:
                        raise Exception('Axis names do not match')
                    if old['axis_types'] != value['axis_types']:
                        raise Exception('Axis types do not match')
                    if old['coordinate_reference_system'] != value['coordinate_reference_system']:
                        raise Exception('Coordinate reference system does not match')

                    if old['axis_offsets'] != value['axis_offsets']:
                        if highest_resolution:
                            if abs(old['axis_offsets'][0]) < abs(value['axis_offsets'][0]):
                                retrieved_metadata[key] = value
                        else:
                            raise Exception("Axis offsets do not match")

                if key == 'size' and key in retrieved_metadata:
                    old = retrieved_metadata[key]
                    if old > value:
                        retrieved_metadata[key] = value

                retrieved_metadata.setdefault(key, value)
        ds = None



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from django.conf import settings
from django.utils.module_loading import import_string

from eoxserver.services.ows.wcs.config import (
    DEFAULT_EOXS_CAPABILITIES_RENDERERS,
    DEFAULT_EOXS_COVERAGE_DESCRIPTION_RENDERERS,
    DEFAULT_EOXS_COVERAGE_RENDERERS,
)

COVERAGE_RENDERERS = None
COVERAGE_DESCRIPTION_RENDERERS = None
CAPABILITIES_RENDERERS = None


def _setup_capabilities_renderers():
    global CAPABILITIES_RENDERERS
    specifiers = getattr(
        settings, 'EOXS_CAPABILITIES_RENDERERS',
        DEFAULT_EOXS_CAPABILITIES_RENDERERS
    )
    CAPABILITIES_RENDERERS = [
        import_string(identifier)()
        for identifier in specifiers
    ]


def _setup_coverage_description_renderers():
    global COVERAGE_DESCRIPTION_RENDERERS
    specifiers = getattr(
        settings, 'EOXS_COVERAGE_RENDERERS',
        DEFAULT_EOXS_COVERAGE_DESCRIPTION_RENDERERS
    )
    COVERAGE_DESCRIPTION_RENDERERS = [
        import_string(identifier)()
        for identifier in specifiers
    ]


def _setup_coverage_renderers():
    global COVERAGE_RENDERERS
    specifiers = getattr(
        settings, 'EOXS_COVERAGE_RENDERERS',
        DEFAULT_EOXS_COVERAGE_RENDERERS
    )
    COVERAGE_RENDERERS = [
        import_string(specifier)()
        for specifier in specifiers
    ]


def get_capabilities_renderer(params):
    if not CAPABILITIES_RENDERERS:
        _setup_capabilities_renderers()

    for renderer in CAPABILITIES_RENDERERS:
        if renderer.supports(params):
            return renderer
    return None


def get_coverage_description_renderer(params):
    if not COVERAGE_DESCRIPTION_RENDERERS:
        _setup_coverage_description_renderers()

    for renderer in COVERAGE_DESCRIPTION_RENDERERS:
        if renderer.supports(params):
            return renderer
    return None


def get_coverage_renderer(params):
    if not COVERAGE_RENDERERS:
        _setup_coverage_renderers()

    for renderer in COVERAGE_RENDERERS:
        if renderer.supports(params):
            return renderer
    return None



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from eoxserver.contrib import mapserver as ms
from eoxserver.services.mapserver.wms.layerfactories.base import (
    AbstractLayerFactory, BaseStyleMixIn, PolygonLayerMixIn
)


class CoverageMaskedOutlinesLayerFactory(BaseStyleMixIn, PolygonLayerMixIn,
                                         AbstractLayerFactory):
    suffixes = ("_masked_outlines",)
    requires_connection = False

    def generate(self, eo_object, group_layer, suffix, options):
        # don't generate any layers, but add the footprint as feature to the
        # group layer

        if group_layer:
            layer = group_layer
        else:
            layer = self._create_polygon_layer(
                eo_object.identifier + "_masked_outlines"
            )

        coverage = eo_object.cast()
        layer.setMetaData("eoxs_geometry_reversed", "true")

        mask_items = coverage.data_items.filter(
            semantic__startswith="polygonmask"
        )

        # use the whole footprint if no mask is given
        if not len(mask_items):
            shape = ms.shapeObj.fromWKT(eo_object.footprint.wkt)
            shape.initValues(1)
            shape.setValue(0, eo_object.identifier)
            layer.addFeature(shape)

        yield layer, mask_items

    def generate_group(self, name):
        return self._create_polygon_layer(name)



#-------------------------------------------------------------------------------
#
#  WPS Literal Data - data-types - unit-tests
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------
# pylint: disable=missing-docstring, line-too-long

from unittest import TestCase, main
from datetime import time, date, datetime, timedelta
from eoxserver.services.ows.wps.parameters import (
    Boolean, Integer, Double, String,
    Duration, Date, Time, DateTime, DateTimeTZAware, CRSType,
)

#------------------------------------------------------------------------------


class BaseTestMixin(object):
    # pylint: disable=invalid-name
    def testGeneral(self):
        self.assertTrue(self.name == self.dtype.name)
        self.assertTrue(self.dtype_diff is self.dtype.get_diff_dtype())

    def testParseOK(self):
        for src, dst in self.parsed:
            try:
                res = self.dtype.parse(src)
            except:
                print ("\n%s: input: %r" % (type(self).__name__, src))
                raise

            try:
                self.assertTrue(isinstance(res, type(dst)))
                self.assertTrue(res == dst or (res != res and dst != dst))
            except:
                print ("\n%s: %r != %r" % (type(self).__name__, res, dst))
                raise

    def testEncodeOK(self):
        for src, dst in self.encoded:
            try:
                res = self.dtype.encode(src)
            except:
                print ("\n%s: input: %r" % (type(self).__name__, src))
                raise
            try:
                self.assertTrue(isinstance(res, type(dst)))
                self.assertTrue(res == dst)
            except:
                print ("\n%s: %r != %r" % (type(self).__name__, res, dst))
                raise

    def testParseFail(self):
        for src in self.parsed_rejected:
            self.assertRaises(ValueError, self.dtype.parse, src)

    def testEncodeFail(self):
        for src in self.encoded_rejected:
            self.assertRaises(ValueError, self.dtype.encode, src)


class TimeZoneTestMixin(object):
    # pylint: disable=invalid-name, too-few-public-methods
    def testParseTimeZone(self):
        for src, dst in self.parsed:
            try:
                res = self.dtype.parse(src)
            except:
                print ("\n input: %r" % src)
                raise
            try:
                if dst.tzinfo is None:
                    self.assertTrue(res.tzinfo is None)
                else:
                    self.assertTrue(res.tzinfo is not None)
                    self.assertTrue(res.utcoffset() == dst.utcoffset())
            except:
                print ("\n%r != %r" % (res, dst))
                raise

#------------------------------------------------------------------------------

class TestDataTypeBool(TestCase, BaseTestMixin):
    def setUp(self):
        self.name = 'boolean'
        self.dtype = Boolean
        self.dtype_diff = self.dtype
        self.encoded = [
            (True, u'true'),
            (1, u'true'),
            ('Anything', u'true'),
            (False, u'false'),
            (0, u'false'),
            (None, u'false'),
            ([], u'false'),
        ]
        self.encoded_rejected = []
        self.parsed = [
            ('true', True),
            ('1', True),
            ('false', False),
            ('0', False),
            (True, True),
            (self, True),
            (False, False),
            (None, False),
            ([], False),
        ]
        self.parsed_rejected = [
            'string',
            u'unicode',
        ]


class TestDataTypeInt(TestCase, BaseTestMixin):
    def setUp(self):
        self.name = 'integer'
        self.dtype = Integer
        self.dtype_diff = self.dtype
        self.encoded = [
            (1, u'1'),
            (-1, u'-1'),
            (False, u'0'),
            (True, u'1'),
            (0xFFFFFFFFFFFFFFFFFF, u'4722366482869645213695'),
            (-0xFFFFFFFFFFFFFFFFFF, u'-4722366482869645213695'),
        ]
        self.encoded_rejected = [
            float('NaN'),
            'anything',
        ]

        self.parsed = [
            (u'+0', 0),
            (u'-0', 0),
            ('24', 24),
            ('32145', 32145),
            (-1, -1),
            (u'4722366482869645213695', 0xFFFFFFFFFFFFFFFFFF),
            ('-4722366482869645213695', -4722366482869645213695),
        ]
        self.parsed_rejected = [
            'nan',
            u'-inf',
            u'24anything',
            '2.5'
        ]


class TestDataTypeFloat(TestCase, BaseTestMixin):
    def setUp(self):
        self.name = 'double'
        self.dtype = Double
        self.dtype_diff = self.dtype
        self.encoded = [
            (1e250, u'1e+250'),
            (-1e-250, u'-1e-250'),
            (-12345678.9012345678, u'-12345678.9012346'),
            (0.6666666666666666, u'0.666666666666667'),
            (-0.0, u'-0'),
            (float('-inf'), u'-inf'),
            (float('nan'), u'nan'),
        ]
        self.encoded_rejected = [
            'anything',
        ]
        self.parsed = [
            (u'1e250', 1e+250),
            ('-1e-250', -1e-250),
            ('16.25', 16.25),
            ('-inf', float('-inf')),
            ('nan', float('nan')),
        ]
        self.parsed_rejected = [
            u'24anything',
        ]


class TestDataTypeString(TestCase, BaseTestMixin):
    def setUp(self):
        sample_unicode = u'P\u0159\xedli\u0161\u017elu\u0165ou\u010dk\xfd k' \
                      u'\u016f\u0148 \xfap\u011bl\u010f\xe1belsk\xe9 \xf3dy.'
        self.name = 'string'
        self.dtype = String
        self.dtype_diff = None
        self.encoded = [
            ('TEST', u'TEST'),
            (sample_unicode, sample_unicode),
        ]
        self.encoded_rejected = []
        self.parsed = [
            (sample_unicode, sample_unicode),
        ]
        self.parsed_rejected = []


class TestDataTypeDuration(TestCase, BaseTestMixin):
    def setUp(self):
        self.name = 'duration'
        self.dtype = Duration
        self.dtype_diff = self.dtype
        self.encoded = [
            (timedelta(2, 11911, 654321), u'P2DT3H18M31.654321S'),
            (-timedelta(2, 11911, 654321), u'-P2DT3H18M31.654321S'),
            (timedelta(2, 11911, 0), u'P2DT3H18M31S'),
            (-timedelta(2, 11911, 0), u'-P2DT3H18M31S'),
            (timedelta(2, 0, 654321), u'P2DT0.654321S'),
            (timedelta(2, 0, 654321), u'P2DT0.654321S'),
            (-timedelta(561, 0, 0), u'-P561D'),
            (-timedelta(561, 0, 0), u'-P561D'),
            (timedelta(0, 11911, 654321), u'PT3H18M31.654321S'),
            (-timedelta(0, 11911, 654321), u'-PT3H18M31.654321S'),
            (timedelta(0, 0, 0), u'PT0S'),
            (-timedelta(0, 0, 0), u'PT0S'),
        ]
        self.encoded_rejected = [
            'anything',
        ]
        self.parsed = [
            (u'P4Y3M2DT3H18M31.654321S', timedelta(1552, 11911, 654321)),
            ('-P4Y3M2DT3H18M31.654321S', -timedelta(1552, 11911, 654321)),
            (u'P1.6Y1.6M1.6DT1.6H1.6M0S', timedelta(633, 57696, 0)),
            (u'-P1.6Y1.6M1.6DT1.6H1.6M0S', -timedelta(633, 57696, 0)),
            (u'PT0S', timedelta(0, 0, 0)),
            (u'P0Y', timedelta(0, 0, 0)),
        ]
        self.parsed_rejected = [
            u'anything',
            u'P1S',
            u'P1H',
            u'PT1Y',
            u'PT1D',
        ]


class TestDataTypeDate(TestCase, BaseTestMixin):
    def setUp(self):
        self.name = 'date'
        self.dtype = Date
        self.dtype_diff = Duration
        self.encoded = [
            (date(1830, 6, 7), u'1830-06-07'),
            (date(2014, 3, 31), u'2014-03-31'),
        ]
        self.encoded_rejected = [
            'anything',
        ]
        self.parsed = [
            (date(1830, 6, 7), date(1830, 6, 7)),
            (u'1830-06-07', date(1830, 6, 7)),
            ('2014-03-31', date(2014, 3, 31)),
        ]
        self.parsed_rejected = [
            u'anything',
            u'2014-02-29',
            u'2014-13-01',
            u'2014-02-00',
            u'2014-00-01'
        ]


class TestDataTypeTime(TestCase, BaseTestMixin):
    # TODO: time-zones
    def setUp(self):
        self.name = 'time'
        self.dtype = Time
        self.dtype_diff = Duration
        self.encoded = [
            (time(0, 0, 0, 0), u'00:00:00'),
            (time(12, 30, 30, 500000), u'12:30:30.500000'),
            (time(23, 59, 59, 999999), u'23:59:59.999999'),
        ]
        self.encoded_rejected = [
            'anything',
        ]
        self.parsed = [
            (time(12, 30, 30, 500000), time(12, 30, 30, 500000)),
            (u'00:00:00Z', time(0, 0, 0, 0)),
            ('12:30:30.5', time(12, 30, 30, 500000)),
            ('23:59:59.999999+01:30', time(23, 59, 59, 999999)),
        ]
        self.parsed_rejected = [
            u'anything',
            '24:00:00',
            '18:60:00',
            '18:30:60',
        ]


class TestDataTypeDateTime(TestCase, BaseTestMixin, TimeZoneTestMixin):
    def setUp(self):
        self.name = 'dateTime'
        self.dtype = DateTime
        self.dtype_diff = Duration
        self.encoded = [
            (
                datetime(2014, 6, 1, 12, 30, 14, 123456, DateTime.UTC),
                u'2014-06-01T12:30:14.123456Z'
            ),
            (
                datetime(2014, 6, 1, 12, 30, 14, 500000, DateTime.TZOffset(90)),
                u'2014-06-01T12:30:14.500000+01:30'
            ),
            (
                datetime(2014, 6, 1, 12, 30, 0, 0, DateTime.UTC),
                u'2014-06-01T12:30:00Z'
            ),
            (
                datetime(2014, 6, 1, 12, 30, 0, 0),
                u'2014-06-01T12:30:00'
            ),
        ]
        self.encoded_rejected = [
            'anything'
        ]
        self.parsed = [
            (
                datetime(2014, 6, 1, 11, 00, 14, 123456),
                datetime(2014, 6, 1, 11, 00, 14, 123456)
            ),
            (
                datetime(2014, 6, 1, 11, 00, 14, 123456, DateTime.TZOffset(90)),
                datetime(2014, 6, 1, 11, 00, 14, 123456, DateTime.TZOffset(90))
            ),
            (
                datetime(2014, 6, 1, 11, 00, 14, 123456, DateTime.UTC),
                datetime(2014, 6, 1, 11, 00, 14, 123456, DateTime.UTC)),
            (
                u'2014-06-01T12:30:14.123456',
                datetime(2014, 6, 1, 12, 30, 14, 123456)),
            (
                u'2014-06-01T12:30:14.123456Z',
                datetime(2014, 6, 1, 12, 30, 14, 123456, DateTime.UTC)
            ),
            (
                u'2014-06-01T12:30:14.123456+01:30',
                datetime(2014, 6, 1, 12, 30, 14, 123456, DateTime.TZOffset(90))
            ),
            (
                u'2014-06-01 12:30:14',
                datetime(2014, 6, 1, 12, 30, 14, 0)
            ),
            (
                u'2014-06-01T00:00Z',
                datetime(2014, 6, 1, 0, 0, 0, 0, DateTime.UTC)
            ),
        ]
        self.parsed_rejected = [
            u'anything',
            u'2014-06-01T12:30:60',
            u'2014-06-01T12:60:30',
            u'2014-06-01T24:00:00',
            u'2014-02-29T00:00',
            u'2014-13-01T00:00',
            u'2014-02-00T00:00',
            u'2014-00-01T00:00',
        ]


class TestDataTypeDateTimeTZAware(TestCase, BaseTestMixin, TimeZoneTestMixin):
    def setUp(self):
        self.name = 'dateTime'
        self.dtype = DateTimeTZAware(DateTime.TZOffset(90))
        self.dtype_diff = Duration
        self.encoded = [
            (
                datetime(2014, 6, 1, 12, 30, 14, 123456),
                u'2014-06-01T12:30:14.123456+01:30'
            ),
            (
                datetime(2014, 6, 1, 12, 30, 14, 123456, DateTime.TZOffset(-90)),
                u'2014-06-01T12:30:14.123456-01:30'
            ),
            (
                datetime(2014, 6, 1, 12, 30, 14, 123456, DateTime.UTC),
                u'2014-06-01T12:30:14.123456Z'
            ),
        ]
        self.encoded_rejected = []
        self.parsed = [
            (
                datetime(2014, 6, 1, 12, 30, 14, 123456, DateTime.TZOffset(-90)),
                datetime(2014, 6, 1, 12, 30, 14, 123456, DateTime.TZOffset(-90))
            ),
            (
                datetime(2014, 6, 1, 12, 30, 14, 123456),
                datetime(2014, 6, 1, 12, 30, 14, 123456, DateTime.TZOffset(90))
            ),
            (
                datetime(2014, 6, 1, 12, 30, 14, 123456, DateTime.UTC),
                datetime(2014, 6, 1, 12, 30, 14, 123456, DateTime.UTC)
            ),
            (
                u'2014-06-01T12:30:14.123456-01:30',
                datetime(2014, 6, 1, 12, 30, 14, 123456, DateTime.TZOffset(-90))
            ),
            (
                u'2014-06-01 12:30:14',
                datetime(2014, 6, 1, 12, 30, 14, 0, DateTime.TZOffset(90))
            ),
            (
                u'2014-06-01T00:00Z',
                datetime(2014, 6, 1, 0, 0, 0, 0, DateTime.UTC)
            ),
        ]
        self.parsed_rejected = []


class TestDataTypeDateTimeTZAwareWithTZConversion(TestCase, BaseTestMixin, TimeZoneTestMixin):
    def setUp(self):
        self.name = 'dateTime'
        self.dtype = DateTimeTZAware(DateTime.TZOffset(90), DateTime.TZOffset(-120))
        self.dtype_diff = Duration
        self.encoded = [
            (
                datetime(2014, 6, 1, 12, 30, 14, 123456),
                u'2014-06-01T09:00:14.123456-02:00'
            ),
            (
                datetime(2014, 6, 1, 12, 30, 14, 123456, DateTime.TZOffset(-90)),
                u'2014-06-01T12:00:14.123456-02:00'
            ),
            (
                datetime(2014, 6, 1, 12, 30, 14, 123456, DateTime.UTC),
                u'2014-06-01T10:30:14.123456-02:00'
            ),
        ]
        self.encoded_rejected = []
        self.parsed = [
            (
                datetime(2014, 6, 1, 12, 30, 14, 123456, DateTime.TZOffset(-90)),
                datetime(2014, 6, 1, 12, 0, 14, 123456, DateTime.TZOffset(-120))
            ),
            (
                datetime(2014, 6, 1, 12, 30, 14, 123456),
                datetime(2014, 6, 1, 9, 0, 14, 123456, DateTime.TZOffset(-120))
            ),
            (
                datetime(2014, 6, 1, 12, 30, 14, 123456, DateTime.UTC),
                datetime(2014, 6, 1, 10, 30, 14, 123456, DateTime.TZOffset(-120))
            ),
            (
                u'2014-06-01T12:30:14.123456-01:30',
                datetime(2014, 6, 1, 12, 0, 14, 123456, DateTime.TZOffset(-120))
            ),
            (
                u'2014-06-01 12:30:14',
                datetime(2014, 6, 1, 9, 0, 14, 0, DateTime.TZOffset(-120))
            ),
            (
                u'2014-06-01T00:00Z',
                datetime(2014, 5, 31, 22, 0, 0, 0, DateTime.TZOffset(-120))
            ),
        ]
        self.parsed_rejected = []


class TestDataTypeCRS(TestCase, BaseTestMixin):
    def setUp(self):
        self.name = 'anyURI'
        self.dtype = CRSType
        self.dtype_diff = None
        self.encoded = [
            (0, u'ImageCRS'),
            (4326, u'http://www.opengis.net/def/crs/EPSG/0/4326'),
        ]
        self.encoded_rejected = [
            -1,
        ]
        self.parsed = [
            ('ImageCRS', 0),
            ('EPSG:4326', 4326),
            ('http://www.opengis.net/def/crs/EPSG/0/4326', 4326),
            ('urn:ogc:def:crs:epsg:6.2:4326', 4326)
        ]
        self.parsed_rejected = [
            "anything",
            "EPSG:0"
        ]

#------------------------------------------------------------------------------

if __name__ == '__main__':
    main()



# -----------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# -----------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# -----------------------------------------------------------------------------

from eoxserver.core.decoders import xml, kvp, typelist, lower
from eoxserver.services.ows.wcs.basehandlers import (
    WCSGetCapabilitiesHandlerBase
)
from eoxserver.services.ows.wcs.parameters import WCSCapabilitiesRenderParams
from eoxserver.services.ows.wcs.v10.util import nsmap


class WCS10GetCapabilitiesHandler(WCSGetCapabilitiesHandlerBase):
    versions = ("1.0.0",)
    methods = ['GET', 'POST']

    def get_decoder(self, request):
        if request.method == "GET":
            return WCS10GetCapabilitiesKVPDecoder(request.GET)
        elif request.method == "POST":
            return WCS10GetCapabilitiesXMLDecoder(request.body)

    def get_params(self, coverages, decoder):
        return WCSCapabilitiesRenderParams(
            coverages, "1.0.0", decoder.sections, decoder.acceptlanguages,
            decoder.acceptformats, decoder.updatesequence
        )


class WCS10GetCapabilitiesKVPDecoder(kvp.Decoder):
    sections            = kvp.Parameter(type=typelist(lower, ","), num="?")
    updatesequence      = kvp.Parameter(num="?")
    acceptversions      = kvp.Parameter(type=typelist(str, ","), num="?")
    acceptformats       = kvp.Parameter(type=typelist(str, ","), num="?", default=["text/xml"])
    acceptlanguages     = kvp.Parameter(type=typelist(str, ","), num="?")
    cql                 = kvp.Parameter(num="?")


class WCS10GetCapabilitiesXMLDecoder(xml.Decoder):
    sections            = xml.Parameter("/ows:Sections/ows:Section/text()", num="*")
    updatesequence      = xml.Parameter("/@updateSequence", num="?")
    acceptversions      = xml.Parameter("/ows:AcceptVersions/ows:Version/text()", num="*")
    acceptformats       = xml.Parameter("/ows:AcceptFormats/ows:OutputFormat/text()", num="*", default=["text/xml"])
    acceptlanguages     = xml.Parameter("/ows:AcceptLanguages/ows:Language/text()", num="*")

    # TODO: find suitable place in XML to pass CQL queries
    cql                 = None

    namespaces = nsmap



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from eoxserver.core import Component, implements
from eoxserver.core.decoders import xml, kvp
from eoxserver.services.ows.interfaces import (
    ServiceHandlerInterface, GetServiceHandlerInterface, 
    PostServiceHandlerInterface
)
from eoxserver.services.ows.wcs.basehandlers import WCSGetCoverageHandlerBase
from eoxserver.services.ows.wcs.v10.parameters import WCS10CoverageRenderParams


class WCS10GetCoverageHandler(WCSGetCoverageHandlerBase, Component):
    implements(ServiceHandlerInterface)
    implements(GetServiceHandlerInterface)
    #implements(PostServiceHandlerInterface)

    versions = ("1.0.0",)
    methods = ['GET']

    def get_decoder(self, request):
        if request.method == "GET":
            return WCS10GetCoverageKVPDecoder(request.GET)
        # TODO: implement POST
        elif request.method == "POST":
            #return WCS10GetCoverageXMLDecoder(request.body)
            pass


    def get_params(self, coverage, decoder, request):
        return WCS10CoverageRenderParams(
            coverage, decoder.bbox, decoder.crs, decoder.format, 
            decoder.response_crs, decoder.width, decoder.height,
            decoder.resx, decoder.resy, decoder.interpolation
        )


def parse_bbox_kvp(string):
    return map(float, string.split(","))


class WCS10GetCoverageKVPDecoder(kvp.Decoder):
    coverage_id = kvp.Parameter("coverage", num=1)
    crs         = kvp.Parameter(num=1)
    response_crs = kvp.Parameter(num="?")
    bbox        = kvp.Parameter(type=parse_bbox_kvp)
    width       = kvp.Parameter(type=int, num="?")
    height      = kvp.Parameter(type=int, num="?")
    resx        = kvp.Parameter(type=float, num="?")
    resy        = kvp.Parameter(type=float, num="?")
    format      = kvp.Parameter(num=1)
    interpolation = kvp.Parameter(num="?")
    exceptions  = kvp.Parameter(num="?")



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from pprint import pprint

from django.core.management.base import CommandError, BaseCommand
from django.db import transaction
from django.db.models import Q

from eoxserver.core.util.timetools import parse_iso8601
from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.management.commands import (
    CommandOutputMixIn, SubParserMixIn
)
from eoxserver.resources.coverages.registration.registrators.gdal import (
    GDALRegistrator
)


class Command(CommandOutputMixIn, SubParserMixIn, BaseCommand):
    """ Command to manage coverages. This command uses sub-commands for the
        specific tasks: register, deregister
    """
    def add_arguments(self, parser):
        register_parser = self.add_subparser(parser, 'register')
        deregister_parser = self.add_subparser(parser, 'deregister')

        register_parser.add_argument(
            "--data", "--data-location", "-d",
            dest="data_locations", nargs="+", action="append", default=[],
            help=(
                "Add a data location to the coverage. In the form "
                "[[... storage] storage] path"
            )
        )
        register_parser.add_argument(
            "--meta-data", "--meta-data-location", "-m",
            dest="metadata_locations", nargs="+", action="append", default=[],
            help=(
                "Add a meta-data file to the coverage. In the form "
                "[[... storage] storage] path"
            )
        )
        register_parser.add_argument(
            '--type', '--coverage-type', '-t',
            dest='coverage_type_name', default=None,
            help=(
                'The name of the coverage type to associate the coverage with.'
            )
        )
        register_parser.add_argument(
            '--grid', '-g',
            dest='grid', default=None,
            help='The name of the grid to associate the coverage with.'
        )
        register_parser.add_argument(
            "--size", "-s",
            dest="size", default=None, nargs="+",
            help="Override size."
        )
        register_parser.add_argument(
            "--origin", "-o", dest="origin", default=None, nargs="+",
            help="Override origin."
        )
        register_parser.add_argument(
            "--highest-resolution",
            dest="highest_resolution", action="store_true", default=False,
            help=(
                "Optional. If the coverage is comprised of raster files of "
                "different resolutions, the highest resolution is used, and "
                "all raster files of a lower resolution will be resampled."
            )
        )
        register_parser.add_argument(
            "--footprint", "-f",
            dest="footprint", default=None,
            help=(
                "Override footprint. Must be supplied as WKT Polygons or "
                "MultiPolygons."
            )
        )
        register_parser.add_argument(
            "--footprint-from-extent",
            dest="footprint_from_extent", action="store_true", default=False,
            help=(
                "Create the footprint from the coverages extent, reprojected "
                "to WGS 84"
            )
        )
        register_parser.add_argument(
            '--simplify-footprint',
            dest='simplify_footprint_tolerance', nargs='?',
            default=None, type=float,
            help=(
                'Simplify the footprint. Optionally specify a tolerance value.'
            )
        )
        register_parser.add_argument(
            "--identifier", "-i",
            dest="identifier", default=None,
            help="Override identifier."
        )
        register_parser.add_argument(
            "--identifier-template",
            dest="identifier_template", default=None,
            help="Add a template to construct the final identifier."
        )
        register_parser.add_argument(
            "--begin-time", "-b",
            dest="begin_time", default=None, type=parse_iso8601,
            help="Override begin time. Format is ISO8601 datetime strings."
        )
        register_parser.add_argument(
            "--end-time", "-e",
            dest="end_time", default=None, type=parse_iso8601,
            help="Override end time. Format is ISO8601 datetime strings."
        )
        register_parser.add_argument(
            "--product", "--product-identifier", "-p",
            dest="product_identifier", default=None,
            help="Add the coverage to the specified product."
        )
        register_parser.add_argument(
            "--collection", "--collection-identifier", "-c",
            dest="collection_identifiers", action="append", default=[],
            help="Add the coverage to the specified collection."
        )
        register_parser.add_argument(
            "--replace", "-r",
            dest="replace", action="store_true", default=False,
            help=(
                "Optional. If the coverage with the given identifier already "
                "exists, replace it. Without this flag, this would result in "
                "an error."
            )
        )
        register_parser.add_argument(
            "--use-subdatasets", "--subdatasets",
            dest="use_subdatasets", action="store_true", default=False,
            help=(
                "Optional. Tell the registrator to interpret colons in the "
                "filename as subdataset specifiers. Default is False."
            )
        )
        register_parser.add_argument(
            '--print-identifier', dest='print_identifier',
            default=False, action='store_true',
            help=(
                'When this flag is set, only the identifier of the registered '
                'coverage will be printed to stdout.'
            )
        )

        deregister_parser.add_argument(
            '--all', '-a', action="store_true",
            default=False, dest='all_coverages',
            help='When this flag is set, all the coverage are selected to be deregisterd'
        )


        deregister_parser.add_argument(
                'identifier', default=None, nargs='?',
                help='The identifier of the coverage to deregister.'
            )

        deregister_parser.add_argument(
            '--not-refresh-collections', dest='not_refresh_collections',
            default=False, action='store_true',
            help=(
                'When this flag is set, the collections and mosaics this '
                'coverage is part of will not have their metadata refreshed.'
            )
        )

    @transaction.atomic
    def handle(self, subcommand, *args, **kwargs):
        """ Dispatch sub-commands: register, deregister.
        """
        if subcommand == "register":
            self.handle_register(*args, **kwargs)
        elif subcommand == "deregister":
            self.handle_deregister(*args, **kwargs)

    def handle_register(self, coverage_type_name,
                        data_locations, metadata_locations,
                        **kwargs):
        """ Handle the creation of a new coverage.
        """
        overrides = {
            key: kwargs[key]
            for key in [
                'begin_time', 'end_time', 'footprint', 'identifier',
                'origin', 'size', 'grid'
            ]
            if kwargs.get(key)
        }

        report = GDALRegistrator().register(
            data_locations=data_locations,
            metadata_locations=metadata_locations,
            coverage_type_name=coverage_type_name,
            footprint_from_extent=kwargs['footprint_from_extent'],
            overrides=overrides,
            identifier_template=kwargs['identifier_template'],
            highest_resolution=kwargs['highest_resolution'],
            replace=kwargs['replace'],
            use_subdatasets=kwargs['use_subdatasets'],
            simplify_footprint_tolerance=kwargs.get(
                'simplify_footprint_tolerance'
            ),
        )

        product_identifier = kwargs['product_identifier']
        if product_identifier:
            product_identifier = product_identifier.format(
                identifier=report.coverage.identifier
            )
            try:
                product = models.Product.objects.get(
                    identifier=product_identifier
                )
            except models.Product.DoesNotExist:
                raise CommandError('No such product %r' % product_identifier)
            models.product_add_coverage(product, report.coverage)

        for collection_identifier in kwargs['collection_identifiers']:
            try:
                collection = models.Collection.objects.get(
                    identifier=collection_identifier
                )
            except models.Collection.DoesNotExist:
                raise CommandError(
                    'No such collection %r' % collection_identifier
                )
            models.collection_insert_eo_object(collection, report.coverage)

        if kwargs['print_identifier']:
            print(report.coverage.identifier)

        else:
            self.print_msg(
                'Successfully registered coverage %s'
                % report.coverage.identifier
            )

            if int(kwargs.get('verbosity', 0)) > 1:
                pprint(report.metadata_parsers)
                pprint(report.retrieved_metadata)

    def handle_deregister(self, identifier, not_refresh_collections, all_coverages, **kwargs):
        """ Handle the deregistration a coverage
        """
        if not all_coverages and not identifier:
            raise CommandError('please specify a coverage/s to remove')
        else:
            if all_coverages:
                coverages = models.Coverage.objects.all()
            elif identifier:
                coverages = [models.Coverage.objects.get(identifier=identifier)]
            for coverage in coverages:
                try:
                    collections = list(coverage.collections.all())
                    mosaics = list(coverage.mosaics.all())
                    grid = coverage.grid

                    if not not_refresh_collections:
                        for collection in collections:
                            models.collection_exclude_eo_object(collection, coverage)

                        for mosaic in mosaics:
                            models.mosaic_exclude_coverage(mosaic, coverage)
                    coverage_id = coverage.identifier
                    coverage.delete()

                    grid_used = models.EOObject.objects.filter(
                        Q(coverage__grid=grid) | Q(mosaic__grid=grid),
                    ).exists()

                    # clean up grid as well, if it is not referenced anymore
                    # but saving named (user defined) grids
                    if grid and not grid.name and not grid_used:
                        grid.delete()

                    self.print_msg(
                        'Successfully deregistered coverage %s' % coverage_id
                    )

                except models.Coverage.DoesNotExist:
                    raise CommandError('No such Coverage %r' % identifier)
